"""This module contains the abstract class :class:`Muscle` that
is used to implement the acting part of an agent.

"""
import logging
import signal
import uuid
from abc import ABC, abstractmethod

import setproctitle
import zmq
import zmq.asyncio

from palaestrai.core import MajorDomoWorker
from palaestrai.core.protocol import (
    AgentShutdownRequest,
    AgentShutdownResponse,
    AgentUpdateRequest,
    AgentUpdateResponse,
    EnvironmentResetNotificationRequest,
    EnvironmentResetNotificationResponse,
    MuscleShutdownRequest,
    MuscleUpdateRequest,
)
from palaestrai.core.serialisation import deserialize, serialize
from ..types.mode import Mode
from ..util.exception import UnknownModeError

LOG = logging.getLogger(__name__)


class Muscle(ABC):
    """An acting entity in an environment.

    Each Muscle is an acting entity in an environment: Given a sensor input,
    it proposes actions. Thus, Muscles implement input-to-action mappings.
    A muscle does, however, not learn by itself; for that, it needs a
    :class:`~Brain`. Every time a muscle acts, it sends the following inputs
    to a Brain:

    * Sensor inputs it received
    * actuator set points it provided
    * reward received from the proposed action.

    When implementing an algorithm, you have to derive from the Muscle ABC and
    provide the following methods:

    1. :func:`~propose_actions`, which implements the input-to-action mapping
    2. :func:`~update`, which handles how updates from the :class:`~Brain` are
       incorporated into the muscle.

    Parameters
    ----------
    broker_connection : str
        the URI which is used to connect to the simulation broker. It is
        used to communicate with the simulation controller.
    brain_connection : str
        a connection for communication with the brain.
    uid : uuid4
        a universal id, that is either provided or assigned here.
    brain_id: str
        the ID of the brain this muscle belongs to.

    """

    def __init__(self, broker_connection, brain_connection, uid, brain_id):
        self.broker_connection = broker_connection
        self.brain_connection = brain_connection
        self._ctx = None
        self._sync_ctx = None
        self._dealer_socket = None
        self.uid = uid or uuid.uuid4()
        self.brain_id = brain_id
        self._worker = None
        self.run_id = None
        self._last_actions = (None, None)
        self._mode = Mode.TRAIN

    @property
    def context(self):
        """Return the asynchronous zmq context.

        The context will be created if necessary.
        """
        if self._ctx is None:
            self._ctx = zmq.asyncio.Context()
        return self._ctx

    @property
    def sync_context(self):
        """Return the synchronous zmq context.

        The context will be created if necessary.
        """

        if self._sync_ctx is None:
            self._sync_ctx = zmq.Context()
        return self._sync_ctx

    @property
    def worker(self):
        """Return the major domo worker.

        The worker will be created if necessary.
        """

        if self._worker is None:
            self._worker = MajorDomoWorker(self.broker_connection, self.uid)
        return self._worker

    @property
    def dealer_socket(self):
        """Return the zmq dealer socket.

        The socket will be created if necessary.
        """

        if self._dealer_socket is None:
            self._dealer_socket = self.sync_context.socket(zmq.DEALER)
            self._dealer_socket.identity = str(self.uid).encode("ascii")
            self._dealer_socket.connect(self.brain_connection)
        return self._dealer_socket

    def _handle_sigintterm(self, signum, frame):
        LOG.info(
            "Muscle %s(id=0x%x, uid=%s) interrupted by signal %s in frame %s.",
            self.__class__,
            id(self),
            self.uid,
            signum,
            frame,
        )
        raise SystemExit()

    def send_to_brain(self, message, flags=0):
        """
        This method is used for communication with the brain. It is
        needed to update the muscle.

        Parameters
        ----------
        message : MuscleUpdateRequest
            The message to be sent to the brain
        flags : int, optional
            Flags for the socket's send method.

        Returns
        -------
        MuscleUpdateResponse
            Response received from the brain.

        """
        z = serialize(message)
        self.dealer_socket.send(z, flags=flags)
        z = self.dealer_socket.recv_multipart(flags=flags)
        response = deserialize(z)
        return response

    async def run(self):
        """Start the main loop of the muscle.

        This method is handling incoming messages and calls the
        corresponding method.

        If an ´AgentUpdateRequest´ is received, it's processed to a
        ´MuscleUpdateRequest´, which triggers a reaction from the
        ´brain´ module (i.e. ´MuscleUpdateResponse´) and finally an
        ´AgentUpdateResponse´ with the action proposals is sent.

        If an ´AgentShutdownRequest´ is received it's processed to a
        ´MuscleUpdateRequest´ and an ´AgentShutdownResponse´ to initiate
        termination of the agent.

        """
        signal.signal(signal.SIGINT, signal.SIG_DFL)
        signal.signal(signal.SIGABRT, signal.SIG_DFL)
        signal.signal(signal.SIGTERM, signal.SIG_DFL)
        setproctitle.setproctitle("palaestrAI[Muscle-%s]" % self.uid[-6:])

        signal.signal(signal.SIGINT, self._handle_sigintterm)
        signal.signal(signal.SIGTERM, self._handle_sigintterm)

        terminal = False
        reply = None
        LOG.info(
            "Muscle %s(id=0x%x, uid=%s) started: Ready for a power workout.",
            self.__class__,
            id(self),
            self.uid,
        )
        while not terminal:
            request = await self.worker.transceive(reply)

            if request is None:
                raise TypeError
            elif isinstance(request, AgentUpdateRequest):
                reply = self._handle_agent_update(request)
            elif isinstance(request, EnvironmentResetNotificationRequest):
                reply = self.handle_environment_reset_notification(request)
            elif isinstance(request, AgentShutdownRequest):
                reply = self._handle_agent_shutdown(request)
                terminal = True

        await self.worker.transceive(reply, skip_recv=True)

        LOG.info(
            "Muscle %s(id=0x%x, uid=%s) completed shutdown: Now I am sore.",
            self.__class__,
            id(self),
            self.uid,
        )

    def _handle_agent_update(
        self, request: AgentUpdateRequest
    ) -> AgentUpdateResponse:
        """Handle an agent update.

        Every update request to the muscle is forwarded to the brain.
        The brain answers with information that the muscle can use to
        update itself.

        Finally, an update response is prepared.

        Parameters
        ----------
        request : AgentUpdateRequest
            The update request from the simulation controller.
            Contains, among other information, the current sensor
            readings and the reward of the previous actions.

        Returns
        -------
        AgentUpdateResponse
            The update response with the agent's actions.

        """
        LOG.debug(
            "Muscle %s(id=0x%x, uid=%s) received %s.",
            self.__class__,
            id(self),
            self.uid,
            request,
        )
        self.run_id = request.experiment_run_id
        self._mode = request.mode
        if self._mode == Mode.TRAIN:
            msg = MuscleUpdateRequest(
                sender_muscle_id=self.uid,
                receiver_brain_id=self.brain_id,
                agent_id=request.receiver_agent_id,
                experiment_run_id=request.experiment_run_id,
                sensor_readings=request.sensors,
                last_actions=self._last_actions[1],
                reward=request.reward,
                is_terminal=request.is_terminal,
                shutdown=False,
            )
            LOG.debug(
                "Muscle(id=0x%x, uid=%s) sending %s to brain.",
                id(self),
                self.uid,
                msg,
            )
            response = self.send_to_brain(msg)
            LOG.debug(
                "Muscle(id=0x%x, uid=%s) received %s from brain",
                id(self),
                self.uid,
                response,
            )
            if response.is_updated:
                self.update(response.updates)
        elif self._mode == Mode.TEST:
            self.prepare_model()
        else:
            raise UnknownModeError()

        if not request.is_terminal:
            self._last_actions = self.propose_actions(
                request.sensors, request.actuators, False
            )
            return AgentUpdateResponse(
                sender_agent_id=self.uid,
                receiver_simulation_controller_id=request.sender,
                experiment_run_id=request.experiment_run_id,
                actuator_information=self._last_actions[0],
                sensor_information=request.sensors,
            )
        else:
            return AgentUpdateResponse(
                sender_agent_id=self.uid,
                receiver_simulation_controller_id=request.sender,
                experiment_run_id=request.experiment_run_id,
                actuator_information=[],
                sensor_information=[],
            )

    def handle_environment_reset_notification(
        self, request: EnvironmentResetNotificationRequest
    ) -> EnvironmentResetNotificationResponse:
        """Handle notification about environment reset.

        Whenever an environment has finished and a new episode is
        started, a notification is send to the agents. Normally,
        they just acknowledge the reset and go on as usual.

        If the agent should, somehow, react to the reset, this method
        can be overwritten to define that reaction.

        Parameters
        ----------
        request: EnvironmentResetNotificationRequest
            The notification request from the simulation controller.

        Returns
        -------
        EnvironmentResetNotificationResponse
            The response for the simulation controller.

        """
        LOG.info(
            "Muscle %s(id=0x%x, uid=%s) acknowledged environment reset.",
            self.__class__,
            id(self),
            self.uid,
        )
        return EnvironmentResetNotificationResponse(
            receiver_simulation_controller_id=request.sender,
            sender_agent_id=self.uid,
        )

    def _handle_agent_shutdown(
        self, request: AgentShutdownRequest
    ) -> AgentShutdownResponse:
        """Handle agent shutdown.

        The muscle sends a final update request to the brain. The
        response from the brain is ignored.
        Finally, a shutdown response is preprared

        Parameters
        ----------
        request : AgentShutdownRequest
            The shutdown request from the simulation controller. This
            message has no further information that need to be
            processed.

        Returns
        -------
        AgentShutdownResponse
            The shutdown response that confirms the shutdown of the
            muscle.

        """
        LOG.info(
            "Muscle %s(id=0x%x, uid=%s) received AgentShutdownRequest.",
            self.__class__,
            id(self),
            self.uid,
        )
        msg = MuscleShutdownRequest(
            sender_muscle_id=self.uid,
            agent_id=request.agent_id,
            experiment_run_id=request.experiment_run_id,
        )
        _ = self.send_to_brain(msg)

        return AgentShutdownResponse(request.experiment_run_id, self.uid, True)

    @abstractmethod
    def setup(self):
        pass

    @abstractmethod
    def propose_actions(
        self, sensors, actuators_available, is_terminal=False
    ) -> tuple:
        """Process new sensor information and produce actuator
        setpoints.

        This is the abstract method which needs to be implemented.
        It works together with a brain and must be tuned to it
        accordingly.

        This method takes a list of :class:sensor_information and has
        to return a list of actuators (this information is given by the
        actuators_available parameter). How the actuator setpoints are
        produced and how the sensor information are processed is up to
        the developer.

        Mainly it is used for reinforcement learning agents. In this
        case, the muscle has a "finished" model and uses it to generate
        the actuator_information. All deep learning tasks are done by
        the brain. The muscle uses the model exclusively and does not
        make any changes itself.

        Parameters
        ----------
        sensors : list[SensorInformation]
            List of new SensorInformation for all available sensors
        actuators_available : list[ActuatorInformation]
            List of all actuators for which a new setpoint is required.
        is_terminal : bool
            Indicator if the simulation run has terminated

        Returns
        -------
        tuple[list, list]
            A tuple of two lists, the first one will be send to the
            environment and needs to be properly scaled, the second
            one will be send to the brain.
        """
        pass

    @abstractmethod
    def update(self, update):
        """Update the muscle.

        This method is called if the brain sends an update. What is to
        be updated is up to the specific implementation. However, this
        method should update all necessary components.

        """
        pass

    def store_model(self):
        """Store a model persistently.

        This method can be used if the muscle does not use a brain that
        takes care of storing the model.

        """
        pass

    @abstractmethod
    def prepare_model(self):
        """Loading a trained model for testing

        If palaestrai is used to test a trained model it hase to load
        the trained model, either from a Database or from a file.
        Loading from a file should be always implemented so it can be
        used as a fallback solution. Also it could/should be checked
        if the model is already initialized, to reduce execution time.

        """
        pass

    @abstractmethod
    def __repr__(self):
        pass



import enum


class BasicState(enum.Enum):
    """Basis for submodule states

    This enumeration is the basis for most submodules' states. It covers their
    basic live cycle.
    """

    PRISTINE = 0
    INITIALIZING = 1
    INITIALIZED = 2
    RUNNING = 3
    STOPPING = 4
    FINISHED = 5
    CANCELLED = 6
    ERROR = 7
    SIGINT = 8
    SIGABRT = 9
    SIGTERM = 10



import asyncio
import itertools
from typing import TYPE_CHECKING
import io
import traceback
import aiomultiprocess
from . import LOG, RunGovernorState

if TYPE_CHECKING:
    from palaestrai.experiment import RunGovernor


class RGSStoppingRun(RunGovernorState):
    """Represents the STOPPING_RUN state.

    Possible next states are
    * :class:`.RGSFinalizing` in the default case.
    """

    def __init__(self, rgc: "RunGovernor"):
        super().__init__(rgc, "STOPPING_RUN")

    async def run(self):
        await self._stop_processes()

    def next_state(self):
        from . import RGSFinalizing
        from . import RGSStartingSetup

        if self.rgc.experiment_run.has_next_phase(self.rgc.current_phase):
            self.rgc.state = RGSStartingSetup(self.rgc)
        else:
            self.rgc.state = RGSFinalizing(self.rgc)

    async def _stop_processes(self):
        all_processes = [
            p
            for p in itertools.chain(
                self.rgc.sim_controllers.values(),
                self.rgc.env_conductors.values(),
                self.rgc.agent_conductors.values(),
            )
        ]
        try:
            await asyncio.wait(
                [
                    asyncio.create_task(self._join_process(p))
                    for p in all_processes
                ],
                return_when=asyncio.ALL_COMPLETED,
                timeout=15,
            )
        except asyncio.TimeoutError as e:
            LOG.warn(
                "RunGovernor(id=0x%x, uid=%s) "
                "saw a timeout while waiting for all remaining processes "
                "terminate peacefully; will continue to reap: %s",
                id(self.rgc),
                self.rgc.uid,
                e,
            )
        except ValueError:
            LOG.warn(
                "RunGovernor(id=0x%x, uid=%s) received an error during "
                "waiting for tasks to complete.",
                id(self.rgc),
                self.rgc.uid,
            )
            tb = io.StringIO()
            traceback.print_exc(file=tb)
            LOG.warn(
                "RunGovernor(id=0x%x, uid=%s) joining tasks ended "
                " with error %s.",
                id(self.rgc),
                self.rgc.uid,
                tb.getvalue(),
            )
            return

        LOG.info(
            "RunGovernor(id=0x%x, uid=%s) shut down all processes.",
            id(self.rgc),
            self.rgc.uid,
        )

    async def _join_process(self, process: aiomultiprocess.Process):
        LOG.debug(
            "RunGovernor(id=0x%x, uid=%s) "
            "monitoring Process(name=%s, is_alive=%s).",
            id(self.rgc),
            self.rgc.uid,
            process.name,
            process.is_alive(),
        )
        while process.exitcode is None:
            await asyncio.sleep(5)
            LOG.debug(
                "RunGovernor(id=0x%x, uid=%s) waiting "
                "for Process(name=%s, is_alive=%s, exitcode=%s, PID=%s).",
                id(self.rgc),
                self.rgc.uid,
                process.name,
                process.is_alive(),
                process.exitcode,
                process.pid,
            )
        LOG.info(
            "RunGovernor(id=0x%x, uid=%s) saw Process(name=%s) exit "
            "with exitcode=%s.",
            id(self.rgc),
            self.rgc.uid,
            process.name,
            process.exitcode,
        )



import logging

LOG = logging.getLogger(__name__)

from .basic_state import BasicState
from .runtime_config import RuntimeConfig
from .major_domo_client import MajorDomoClient
from .major_domo_broker import MajorDomoBroker
from .major_domo_worker import MajorDomoWorker

__ALL__ = [
    "BasicState",
    "RuntimeConfig",
    "MajorDomoBroker",
    "MajorDomoClient",
    "MajorDomoWorker",
]



import logging

LOG = logging.getLogger(__name__)

from .executor import (
    Executor,
    ExecutorState,
    ExperimentRunStartError,
    ExperimentRunRuntimeInformation,
)
from .experiment_run import ExperimentRun
from .run_governor import RunGovernor
from .termination_condition import TerminationCondition
from .vanilla_rungovernor_termination_condition import (
    VanillaRunGovernorTerminationCondition,
)



class EnvironmentStartRequest:
    def __init__(self, run_id, environment_id):
        self._run_id = run_id
        self._environment_id = environment_id

    @property
    def run_id(self):
        return self._run_id

    @run_id.setter
    def run_id(self, value):
        self._run_id = value

    @property
    def environment_id(self):
        return self._environment_id

    @environment_id.setter
    def environment_id(self, value):
        self._environment_id = value



from . import LOG


class ShutdownResponse:
    def __init__(self, experiment_run_id):
        self._experiment_run_id = experiment_run_id

    @property
    def run_id(self):
        """Deprecated: Use experiment_run_id instead."""
        LOG.debug(
            f"Run_id property deprecated in class {self.__class__}. Use experiment_run_id instead."
        )
        return self.experiment_run_id

    @run_id.setter
    def run_id(self, value):
        """Deprecated: Use experiment_run_id instead."""
        LOG.debug(
            f"Run_id property deprecated in class {self.__class__}. Use experiment_run_id instead."
        )
        self.experiment_run_id(value)

    @property
    def experiment_run_id(self):
        return self._experiment_run_id

    @experiment_run_id.setter
    def experiment_run_id(self, value):
        self._experiment_run_id = value

    def __eq__(self, other):
        if (
            type(other) is type(self)
            and other.experiment_run_id == self.experiment_run_id
        ):
            return True
        return False



#!/usr/bin/env python3

import os
import alembic.config
from palaestrai.core import RuntimeConfig


def main():
    config = RuntimeConfig()
    try:
        config.load()
        if not config.store_uri:
            raise KeyError()
    except KeyError:
        print(
            "Please create a runtime config (one of %s), "
            "and set the 'store_uri' options.\n"
            "My configuration, loaded from %s, does not contain it."
            % (config.CONFIG_FILE_PATHS, config._config_file_path)
        )
        exit(1)
    except FileNotFoundError as e:
        print(
            "Please create a runtime config " "and set the 'store_uri' option."
        )
        print("%s." % e)
        exit(1)
    here = os.path.dirname(os.path.abspath(__file__))
    alembic_args = ["-c", os.path.join(here, "alembic.ini"), "upgrade", "head"]
    alembic.config.main(argv=alembic_args)



from unittest import TestCase

from palaestrai.agent import RewardInformation
from palaestrai.simulation import VanillaSimControllerTerminationCondition
from palaestrai.core.protocol import EnvironmentUpdateResponse
from palaestrai.agent.sensor_information import SensorInformation
from palaestrai.types import Discrete


class TestVanillaSimControllerTerminationCondition(TestCase):
    def test_check_termination_true(self):
        term_cond = VanillaSimControllerTerminationCondition()
        self.assertTrue(
            term_cond.check_termination(
                EnvironmentUpdateResponse(
                    sender_environment_id="0",
                    receiver_simulation_controller_id="0",
                    experiment_run_id="0",
                    environment_conductor_id="0",
                    sensors=[SensorInformation(1, Discrete(1), 0)],
                    reward=[RewardInformation(0.0, Discrete(1), "Test")],
                    is_terminal=True,
                )
            )
        )

    def test_check_termination_False(self):
        term_cond = VanillaSimControllerTerminationCondition()
        self.assertFalse(
            term_cond.check_termination(
                EnvironmentUpdateResponse(
                    sender_environment_id="0",
                    receiver_simulation_controller_id="0",
                    experiment_run_id="0",
                    environment_conductor_id="0",
                    sensors=[SensorInformation(1, Discrete(1), 0)],
                    reward=[RewardInformation(0.0, Discrete(1), "Test")],
                    is_terminal=False,
                )
            )
        )



import itertools
from typing import TYPE_CHECKING

from palaestrai.core.protocol import (
    SimulationStartRequest,
    SimulationStartResponse,
)
from palaestrai.util.exception import InvalidResponseError

from . import LOG, RunGovernorState

if TYPE_CHECKING:
    from palaestrai.experiment import RunGovernor


class RGSStartingRun(RunGovernorState):
    """Represent the STARTING_RUN state of the run governor.

    This state is entered after STARTING_AGENT_CONDUCTORS succeeds.
    During this state, all previously created processes will be
    started, which marks the begin of the experiment run.

    Possible next states are
    * :class:`.RGSTransceiving in the normal case
    * :class:`.RGSErrorHandlingStarting in the error case

    Parameters
    ----------
    rgc: :class:`.RunGovernor`
        The run governor instance that provides the context for this
        state.

    """

    def __init__(self, ctx: "RunGovernor"):
        super().__init__(ctx, "STARTING_RUN")

    async def run(self):
        self._start_processes()
        await self._send_start_request()
        self._check_reply()
        LOG.info(
            "RunGovernor(id=0x%x, uid=%s) started execution of phase "
            "%d: '%s'.",
            id(self.rgc),
            self.rgc.uid,
            self.rgc.current_phase,
            self.rgc.experiment_run.get_phase_name(self.rgc.current_phase),
        )

    def next_state(self):
        from . import RGSErrorHandlingStarting, RGSTransceiving

        if len(self.rgc.errors) > 0:
            self.rgc.state = RGSErrorHandlingStarting(self.rgc)
        else:
            self.rgc.state = RGSTransceiving(self.rgc)

    def _start_processes(self):
        for proc in itertools.chain(
            self.rgc.sim_controllers.values(),
            self.rgc.env_conductors.values(),
            self.rgc.agent_conductors.values(),
        ):
            proc.start()

    async def _send_start_request(self):
        for sc_uid in self.rgc.sim_controllers.keys():
            LOG.debug(
                "RunGovernor(id=0x%x, uid=%s) requesting start of "
                "SimulationController(uid=%s).",
                id(self.rgc),
                self.rgc.uid,
                sc_uid,
            )
            msg = SimulationStartRequest(
                self.rgc.uid, sc_uid, self.rgc.experiment_run_id
            )
            response = await self.rgc.major_domo_client.send(sc_uid, msg)
            if not isinstance(response, SimulationStartResponse):
                self.add_error(
                    InvalidResponseError(
                        SimulationStartResponse.__class__, response.__class__
                    )
                )
            else:
                LOG.debug(
                    "RunGovernor(id=0x%x, uid=%s) received "
                    "SimulationStartResponse(sc_uid=%s).",
                    id(self.rgc),
                    self.rgc.uid,
                    sc_uid,
                )

    def _check_reply(self):
        request = self.rgc.last_request.pop()
        LOG.debug("Last request is %s", request)



import asyncio
from typing import Union, TextIO, Tuple, Any

from palaestrai.core import RuntimeConfig
from palaestrai.experiment import ExperimentRun, Executor, ExecutorState


def execute(
    experiment_run_definition: Union[ExperimentRun, TextIO, str],
    runtime_config: Union[str, TextIO, dict, None] = None,
) -> Tuple[str, ExecutorState]:
    """Provides a single-line command to start an experiment and set a
    runtime configuration

    Parameters
    ----------
    experiment_run_definition: 1. Already set ExperimentRun object
                               2. Any text stream
                               3. The path to a file
        The configuration from which the experiment is loaded.

    runtime_config:            1. Any text stream
                               2. dict
                               3. None
        The Runtime configuration applicable for the run.
        Note that even when no additional source is provided, runtime will load
        a minimal configuration from build-in defaults.

    Returns
    -------
    Experiment.user_id:  Unique experiment ID of type str
    executor state: The state the executor is now in of type ExecutorState
    """
    if runtime_config:
        RuntimeConfig().load(runtime_config)
    else:
        RuntimeConfig().load()

    if not isinstance(experiment_run_definition, ExperimentRun):
        experiment = ExperimentRun.load(experiment_run_definition)
    else:
        experiment = experiment_run_definition

    executor = Executor()
    executor.schedule(experiment)
    executor_final_state = asyncio.run(executor.execute())

    return experiment.user_id, executor_final_state



"""This module contains the :class:`.RunGovernor`. """
import io
import logging
import traceback
import uuid
from typing import TYPE_CHECKING, Dict, Union

from palaestrai.core import MajorDomoClient, MajorDomoWorker

from .rungovernor_states.base_state import RunGovernorState
from .rungovernor_states.rgs_pristine import RGSPristine

if TYPE_CHECKING:
    import aiomultiprocess
    from palaestrai.experiment import TerminationCondition
    from palaestrai.experiment.experiment_run import ExperimentRun


LOG = logging.getLogger(__name__)


class RunGovernor:
    """This class implements the Run-Governor.

    Upon receiving requests from the executor, a RunGovernor instance
    handles a single experiment run by starting it, initialize the
    simulation controllers, the environment and the agent conductors,
    and, finally, shutting the experiment run down.

    The RunGovernor is implemented as state machine and this class
    provides the context for the distinct state classes. A freshly
    initialized RunGovernor waits in the state PRISTINE until the run
    method is called by the executor. See the distinct state classes
    for more information.

    Parameters
    ----------
    broker_connection: str
        The URI representing the connection to the broker for
        communication with the executor.
    termination_condition: TerminationCondition
        The condition that tells the RunGovernor when to stop the
        simulation.
    rungov_uid: str
        The UUID for this RunGovernor is provided by the executor.

    Attributes
    ----------
    uid: str
        The UUID of this RunGovernor
    termination_condition: :class:`.TerminationCondition`
        A reference to the TerminationCondition instance.
    run_broker: :class:`.MajorDomoBroker`
        The broker for the communication with the simulation
        controller, the agents, and the environments.
    experiment_run_id: str
        The UUID of the current experiment run.
    tasks: List[aiomultiprocess.Process]
        A list of tasks the RunGovernor has started and that it has
        to shutdown in the end.
    worker: :class:`.MajorDomoWorker`
        The major domo worker for handling incoming requests
    client: :class:`.MajorDomoClient`
        The major domo client for sending requests to other workers.
    shutdown: bool
        The major kill switch of the RunGovernor. Setting this to false
        will stop the RunGovernor after the current state.
    state: :class:`.RunGovernorState`
        Holds the current state instance. The first state is PRISTINE.
    errors: List[Exception]
        A list that is used to collect errors raised in the states.

    """

    def __init__(self, broker_uri: str, uid: str = None):
        # self._binding_uri = "tcp://*:45306"

        self.uid = uid if uid else "RunGovernor-%s" % str(uuid.uuid4())
        self.broker_uri = broker_uri
        self.major_domo_client: MajorDomoClient
        self.major_domo_worker: MajorDomoWorker

        self.termination_condition: TerminationCondition
        self.experiment_run: Union[ExperimentRun, None] = None
        self.experiment_run_id: Union[str, None] = None
        self.current_phase = 0
        self.current_episode = 0
        self.transceive_task = None
        self.signal_monitor_task = None
        self.signal_received = None
        self.agent_conductors: Dict[str, aiomultiprocess.Process] = dict()
        self.env_conductors: Dict[str, aiomultiprocess.Process] = dict()
        self.sim_controllers: Dict[str, aiomultiprocess.Process] = dict()

        # List is a "workaround" to satisfy the type checking ...
        self.last_request: list = list()
        self.next_reply: list = list()
        self.stopping = False
        self.shutdown = False
        self.state: RunGovernorState = RGSPristine(self)
        self.errors: list = list()
        self.dead_children: list = list()
        # self._simulation_controllers: dict = dict()

    async def run(self):
        """Start the main loop of the run governor.

        In each iteration, the run method of the current state is
        called.

        """
        while not self.shutdown:
            LOG.info(
                "RunGovernor(id=0x%x, uid=%s) now processing state %s.",
                id(self),
                self.uid,
                self.state.name,
            )
            try:
                await self.state.run()
                self.state.next_state()
            except Exception as err:
                LOG.critical(
                    "RunGovernor(id=0x%x, uid=%s) died in disgrace!",
                    id(self),
                    self.uid,
                )

                tb = io.StringIO()
                traceback.print_exc(file=tb)
                LOG.critical(
                    "RunGovernor(id=0x%x, uid=%s) has crashed in state "
                    "%s with error %s",
                    id(self),
                    self.uid,
                    self.state.name,
                    tb.getvalue(),
                )
                raise err
        LOG.info(
            "RunGovernor(id=0x%x, uid=%s) has finished.", id(self), self.uid
        )



"""Add SimulationController uuid to SimulationInstance

Revision ID: e8749417b866
Revises: 728eceef1424
Create Date: 2020-11-27 02:54:41.312500

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "e8749417b866"
down_revision = "728eceef1424"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "simulation_instances",
        sa.Column("uuid", sa.VARCHAR(255), unique=True, index=True),
    )


def downgrade():
    op.drop_column("simulation_instances", "uuid")



"""initial

Revision ID: f29bac335186
Revises: 
Create Date: 2020-11-22 15:24:18.781971

"""
from alembic import op
import sqlalchemy as sa
from palaestrai.store.database_model import Model
from palaestrai.store.database_base import metadata


# revision identifiers, used by Alembic.
revision = "f29bac335186"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###



import unittest
from unittest.mock import AsyncMock, MagicMock, patch
import asyncio
from palaestrai.experiment import RunGovernor
from palaestrai.experiment.rungovernor_states import (
    RGSFinalizing,
    RGSDone,
)


class TestStoppingRun(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.rungov = RunGovernor(None, None)
        self.rgs = RGSFinalizing(self.rungov)
        self.rungov.state = self.rgs

    async def test_run(self):
        self.rgs._shutdown_tasks = AsyncMock()
        # self.rungov.major_domo_worker.transceive = AsyncMock()

        logger = f"{RunGovernor.__module__}".rsplit(".", 1)[0]
        with self.assertLogs(logger, level="DEBUG") as cm:
            await self.rgs.run()

        self.rgs._shutdown_tasks.assert_called_once()
        self.assertIn("completed shutdown", cm.output[0])

    def test_next_state(self):
        self.rgs.next_state()
        self.assertIsInstance(self.rungov.state, RGSDone)

    @patch(f"{RGSFinalizing.__module__}.os")
    async def test_shutdown_tasks(self, mock_os):
        self.rungov.sim_controllers["sim-1"] = 0
        mock_os.killpg = MagicMock()
        RGSFinalizing._reap_process = AsyncMock()
        await self.rgs._shutdown_tasks()

        mock_os.killpg.assert_called_once()
        RGSFinalizing._reap_process.assert_called_once()

    async def test_reap_process(self):
        def raise_timeout():
            raise asyncio.TimeoutError()

        proc = MagicMock()
        proc.is_alive = MagicMock(side_effect=[True, True])
        proc.terminate = MagicMock()
        proc.kill = MagicMock()
        proc.join = AsyncMock(side_effect=lambda timeout: raise_timeout())

        await self.rgs._reap_process(proc)

        proc.terminate.assert_called_once()
        proc.join.assert_called_once()
        proc.kill.assert_called_once()
        self.assertEqual(2, proc.is_alive.call_count)


if __name__ == "__main__":
    unittest.main()



"""This module contains the class :class:`RewardInformation` that
stores all information the agents need about a single reward.

"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from palaestrai.types import Space


class RewardInformation:
    """Stores information about a single reward.

    Once created, a *RewardInformation* object can be called to
    retrieve its value, e.g.,:

        a = RewardInformation(42, some_space)
        a()  # => 42
        a.reward_value  # => 42


    Parameters
    ----------
    reward_value: int float as described in observation_space
        The value of the reward's last reading. The type of this value
        is described by :attr:`observation_space`
    observation_space : :class:`~.Space`
        An instance of a palaestrai space object defining the type of
        the value
    reward_id: str or int, optional
        A unique identifier for this reward. The agents use the ID only
        for assignment of previous values with the same ID. The ID is
        important, if multiple rewards are available and/or the reward
        is a delayed reward.

    """

    def __init__(self, reward_value, observation_space: Space, reward_id=None):
        self.reward_value = reward_value
        self.observation_space = observation_space
        self.reward_id = reward_id

    def __call__(self):
        """Reads the reward"""
        return self.reward_value

    def __add__(self, other):
        return self.reward_value + other

    def __repr__(self):
        return (
            "RewardInformation("
            f"value={self.reward_value}, observation_space="
            f"{repr(self.observation_space)}, reward_id={self.reward_id})"
        )



from typing import TYPE_CHECKING

from palaestrai.util.exception import (
    AgentConductorFailedError,
    EnvConductorFailedError,
    ExperimentAlreadyRunningError,
    ExperimentSetupFailedError,
    InvalidResponseError,
    SimControllerFailedError,
)

from . import LOG, RunGovernorState

if TYPE_CHECKING:
    from palaestrai.experiment import RunGovernor


class RGSErrorHandlingStarting(RunGovernorState):
    """Represent the ERROR_HANDLING_STARTING state of the run
    governor.

    Expected errors:
    * ExperimentAlreadyRunningError during STARTING_EXPERIMENT
    * ExperimentSetupFailedError during STARTING_EXPERIMENT
    * SimControllerFailedError during STARTING_SIM_CONTROLLERS (?)
    * EnvConductorFailedError during STARTING_ENV_CONDUCTORS (?)
    * AgentConductorFailedError during STARTING_AGENT_CONDUCTORS (?)
    * InvalidResponseError during STARTING_RUN

    Possible next states are
    * :class:`.RGSTransceiving` in the ExperimentAlreadyRunningError
        case.
    * :class:`.RGSStoppingTransceiving` in the
        ExperimentSetupFailedError case.
    * :class:`.RGSStoppingSimulation`  in the cases
        SimControllerFailedError, EnvConductorFailedError,
        AgentConductorFailedError, InvalidResponseError.

    """

    def __init__(self, rgc: "RunGovernor"):
        super().__init__(rgc, "ERROR_HANDLING_STARTING")

    async def run(self):
        LOG.info(
            "RunGovernor(id=0x%x, uid=%s) encountered errors during "
            "experiment starting. Handling these errors now.",
            id(self.rgc),
            self.rgc.uid,
        )
        for (error, state) in self.rgc.errors:
            LOG.info(
                "RunGovernor(id=0x%x, uid=%s) encountered %s during "
                "state %s",
                id(self.rgc),
                self.rgc.uid,
                error.__class__,
                state,
            )
            LOG.debug(
                "RunGovernor(id=0x%x, uid=%s) message of the error: %s",
                id(self.rgc),
                self.rgc.uid,
                error,
            )

        assert len(self.rgc.errors) == 1

    def next_state(self):
        from . import (
            RGSStoppingSimulation,
            RGSStoppingTransceiving,
            RGSTransceiving,
        )

        dispatcher = {
            ExperimentAlreadyRunningError: RGSTransceiving,
            ExperimentSetupFailedError: RGSStoppingTransceiving,
            SimControllerFailedError: RGSStoppingSimulation,
            EnvConductorFailedError: RGSStoppingSimulation,
            AgentConductorFailedError: RGSStoppingSimulation,
            InvalidResponseError: RGSStoppingSimulation,
        }

        self.rgc.state = dispatcher[self.rgc.errors[0][0].__class__](self.rgc)
        self.rgc.errors = list()



"""Majordomo Protocol definitions"""
#  This is the version of MDP/Client we implement
C_CLIENT = b"MDPC01"

#  This is the version of MDP/Worker we implement
W_WORKER = b"MDPW01"

#  MDP/Server commands, as strings
W_READY = b"\001"
W_REQUEST = b"\002"
W_REPLY = b"\003"
W_HEARTBEAT = b"\004"
W_DISCONNECT = b"\005"
W_DESTROY = b"\006"

commands = [
    None,
    b"READY",
    b"REQUEST",
    b"REPLY",
    b"HEARTBEAT",
    b"DISCONNECT",
    b"DESTROY",
]

"""
C_CLIENT = b'MDPC01'

#  This is the version of MDP/Worker we implement
W_WORKER = b'MDPW01'

#  MDP/Server commands, as strings
W_READY = b'\001'
W_REQUEST = b'\002'
W_REPLY = b'\003'
W_HEARTBEAT = b'\004'
W_DISCONNECT = b'\005'

commands = [None, b'READY', b'REQUEST', b'REPLY', b'HEARTBEAT', b'DISCONNECT']
"""
# Note, Python3 type "bytes" are essentially what Python2 "str" were,
# but now we have to explicitly mark them as such.  Type "bytes" are
# what PyZMQ expects by default.  Any user code that uses this and
# related modules may need to be updated.  Here are some guidelines:
#
# String literals that make their way into messages or used as socket
# identfiers need to become bytes:
#
#   'foo' -> b'foo'
#
# Multippart messages, originally formed as lists of strings (smsg)
# need to be washed into strings of bytes (bmsg) like:
#
#   bmsg = [one.encode('utf-8') for one in smsg]
#
# A multipart message recived can be reversed
#
#   smsg = [one.decode() for one in bmsg]



from __future__ import annotations
from typing import List, TYPE_CHECKING
from dataclasses import dataclass

if TYPE_CHECKING:
    from palaestrai.agent import (
        SensorInformation,
        ActuatorInformation,
        RewardInformation,
    )


@dataclass
class AgentUpdateResponse:
    """Responds after an agent has acted via its :class:`Muscle`

    * Sender: :class:`Muscle`
    * Receiver: :class:`SimulationController`

    :param sender_agent_id: ID of the sending agent, e.g., a :class:`Muscle`
    :param receiver_simulation_controller_id: ID of the receiving
        :class:`SimulationController`
    :param experiment_run_id: ID of the experiment run we participate in
    :param sensor_information: list of sensor readings
        :class:`SensorInformation`
    :param actuator_information: list of actuator actions via
        :class:`ActuatorInformation`
    """

    sender_agent_id: str
    receiver_simulation_controller_id: str
    experiment_run_id: str
    sensor_information: List[SensorInformation]
    actuator_information: List[ActuatorInformation]

    @property
    def sender(self):
        return self.sender_agent_id

    @property
    def receiver(self):
        return self.receiver

    @property
    def actuators(self):
        return self.actuator_information



import asyncio
import unittest
from unittest.mock import MagicMock, AsyncMock

from palaestrai.core.protocol import (
    SimulationControllerTerminationRequest,
    SimulationControllerTerminationResponse,
)
from palaestrai.experiment import RunGovernor
from palaestrai.experiment.rungovernor_states import (
    RGSHandlingSimControllerTermination,
    RGSStoppingSimulation,
    RGSTransceiving,
)


class TestHandlingSimControllerTermination(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.rungov = RunGovernor(None, None)
        self.rgs = RGSHandlingSimControllerTermination(self.rungov)
        self.rungov.state = self.rgs
        self.req = SimulationControllerTerminationRequest(
            None, None, None, None, None, None
        )

    async def test_run(self):
        def async_return(result):
            f = asyncio.Future()
            f.set_result(result)
            return f

        self.rungov.last_request.append(self.req)
        self.rgs._handle_termination_request = MagicMock(return_value=False)
        self.rgs._prepare_reply = AsyncMock()
        self.rgs._check_episodes = MagicMock(return_value=False)

        await self.rgs.run()

        self.rgs._handle_termination_request.assert_called_with(
            self.req, False
        )
        self.rgs._prepare_reply.assert_called_with(self.req, False, False)

    def test_next_state(self):
        self.rungov.next_reply.append(
            SimulationControllerTerminationResponse(
                sender_run_governor_id=None,
                receiver_simulation_controller_id=None,
                experiment_run_id=None,
                restart=False,
                complete_shutdown=False,
            )
        )
        self.rgs.next_state()

        self.assertIsInstance(self.rungov.state, RGSTransceiving)

    def test_next_state_complete_shutdown(self):
        self.rungov.next_reply.append(
            SimulationControllerTerminationResponse(
                sender_run_governor_id=None,
                receiver_simulation_controller_id=None,
                experiment_run_id=None,
                restart=False,
                complete_shutdown=True,
            )
        )
        self.rgs.next_state()

        self.assertIsInstance(self.rungov.state, RGSStoppingSimulation)

    def test_handle_termination_request(self):
        self.rungov.sim_controllers["sim-1"] = MagicMock()
        self.req.sender_simulation_controller_id = "sim-1"
        self.rungov.termination_condition = MagicMock()
        self.rungov.termination_condition.check_termination = MagicMock(
            return_value=True
        )

        self.rgs._handle_termination_request(self.req, False)

        self.rungov.termination_condition.check_termination.assert_called_with(
            self.req, dict()
        )

    async def test_prepare_reply(self):

        self.req.sender_simulation_controller_id = "sim-1"

        await self.rgs._prepare_reply(self.req, False, False)

        self.assertIsInstance(
            self.rungov.next_reply[0], SimulationControllerTerminationResponse
        )

        self.assertEqual("sim-1", self.rungov.next_reply[0].receiver)

    def test_check_episodes(self):
        self.rungov.experiment_run = MagicMock()
        self.rungov.experiment_run.get_episodes = MagicMock(return_value=2)

        result = self.rgs._check_episodes()

        self.assertTrue(result)

        result = self.rgs._check_episodes()

        self.assertFalse(result)


if __name__ == "__main__":
    unittest.main()



import unittest
from unittest.mock import AsyncMock, MagicMock
from uuid import uuid4

from palaestrai.agent import SensorInformation, ActuatorInformation
from palaestrai.core.protocol import (
    EnvironmentShutdownRequest,
    EnvironmentShutdownResponse,
    EnvironmentStartRequest,
    EnvironmentStartResponse,
    EnvironmentUpdateRequest,
    EnvironmentUpdateResponse,
    EnvironmentResetRequest,
    EnvironmentResetResponse,
)
from palaestrai.environment.dummy_environment import DummyEnvironment
from palaestrai.environment.dummy_reward import DummyReward
from palaestrai.types import Discrete


def check_identifier(information, identifier="testenv"):
    for i in information:
        if identifier not in i.id:
            print(i.id)
            return False
    return True


class TestEnvironment(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.env = DummyEnvironment(
            "test_conn", uuid4(), DummyReward(), 123, {"identifier": "testenv"}
        )
        self.setup_req = EnvironmentStartRequest(
            environment_id="0",
            run_id="1",
        )
        self.update_req = EnvironmentUpdateRequest(
            sender_simulation_controller="2",
            receiver_environment="0",
            experiment_run_id="1",
            environment_conductor_id="3",
            actuators=list(),
        )
        self.shutdown_req = EnvironmentShutdownRequest(
            experiment_run_id="1", environment_id="0"
        )
        self.reset_req = EnvironmentResetRequest("0", "1")
        # self.reset_rsp = EnvironmentResetResponse("0")

    def test_handle_setup(self):
        self.env.start_environment = MagicMock(
            return_value=(
                [SensorInformation(0, Discrete(1), "0")],
                [ActuatorInformation(0, Discrete(1), "0")],
            )
        )
        rsp = self.env._handle_setup(self.setup_req)

        self.env.start_environment.assert_called_once()
        self.assertIsInstance(rsp, EnvironmentStartResponse)
        self.assertTrue(check_identifier(rsp.sensors))
        self.assertTrue(check_identifier(rsp.actuators))

    def test_handle_update(self):
        self.env.update = MagicMock(
            return_value=([SensorInformation(0, Discrete(1), "0")], False)
        )
        rsp = self.env._handle_update(self.update_req)

        self.env.update.assert_called_once()
        self.assertIsInstance(rsp, EnvironmentUpdateResponse)
        self.assertTrue(check_identifier(rsp.sensors))

    def test_handle_shutdown(self):
        self.env.shutdown = MagicMock(return_value=True)
        rsp = self.env._handle_shutdown(self.shutdown_req)

        self.env.shutdown.assert_called_once()
        self.assertIsInstance(rsp, EnvironmentShutdownResponse)

    def test_handle_reset(self):
        self.env.shutdown = MagicMock()
        self.env.start_environment = MagicMock(
            return_value=(
                [SensorInformation(0, Discrete(1), "0")],
                [ActuatorInformation(0, Discrete(1), "0")],
            )
        )
        result = self.env._handle_reset(self.reset_req)

        self.assertIsInstance(result, EnvironmentResetResponse)
        self.assertTrue(check_identifier(result.sensors))
        self.assertTrue(check_identifier(result.actuators))

        self.env.shutdown.assert_called_once()
        self.env.start_environment.assert_called_once()

    async def test_run(self):
        setup_msg = self.setup_req
        update_msg = self.update_req
        shutdown_msg = self.shutdown_req
        self.env.worker.transceive = AsyncMock()
        self.env.worker.transceive.side_effect = [
            setup_msg,
            update_msg,
            update_msg,
            update_msg,
            shutdown_msg,
            shutdown_msg,  # Final message, will not be handled
        ]
        self.env.start_environment = MagicMock(return_value=(list(), list()))
        self.env.update = MagicMock(return_value=(list(), False))
        await self.env.run()

        self.assertEqual(self.env.worker.transceive.call_count, 6)
        self.env.start_environment.assert_called_once()
        self.assertEqual(self.env.update.call_count, 3)


if __name__ == "__main__":
    unittest.main()



import asyncio
import signal
from typing import TYPE_CHECKING

from palaestrai.core import MajorDomoClient, MajorDomoWorker

from . import LOG, RunGovernorState

if TYPE_CHECKING:
    from palaestrai.experiment import RunGovernor


class RGSInitializing(RunGovernorState):
    """Represent the INITIALIZING state of the run governor.

    This state is entered after the PRISTINE state. During this state,
    a signal handler and a signal monitor for graceful termination are
    initialized and the communication is set up.

    Possible next states are
    * :class:`.RGSTransceiving` in the normal case.
    * :class:`.RGSErrorHandlingInitializing in the error case.

    Parameters
    ----------
    rgc: :class:`.RunGovernor`
        The run governor instance that provides the context for this
        state.

    """

    def __init__(self, rgc: "RunGovernor"):
        super().__init__(rgc, "INITIALIZING")

    async def run(self):
        self._init_signal_handler()
        self._init_signal_monitor()
        self._init_communication()

    def next_state(self):
        from . import RGSErrorHandlingInitializing, RGSTransceiving

        if len(self.rgc.errors) > 0:
            self.rgc.state = RGSErrorHandlingInitializing(self.rgc)
        else:
            self.rgc.state = RGSTransceiving(self.rgc)

    def _init_signal_handler(self):
        signals = {signal.SIGABRT, signal.SIGTERM}
        LOG.debug(
            "RunGovernor(id=0x%x, uid=%s) "
            "registering signal handlers for signals %s.",
            id(self.rgc),
            self.rgc.uid,
            signals,
        )
        loop = asyncio.get_running_loop()
        for signum in signals:
            loop.add_signal_handler(
                signum, self._handle_signal_termination, signum
            )
        loop.add_signal_handler(signal.SIGINT, self._handle_signal_interrupt)

    def _handle_signal_termination(self, signum):
        LOG.info(
            "RunGovernor(id=0x%x, uid=%s) interrupted by signal %s.",
            id(self.rgc),
            self.rgc.uid,
            signum,
        )
        self.rgc.signal_received = signum

    def _handle_signal_interrupt(self):
        LOG.debug(
            "RunGovernor(id=0x%x, uid=%s) received SIGINT; "
            "trusting the upper brass is handling this.",
            id(self.rgc),
            self.rgc.uid,
        )

    def _init_signal_monitor(self):
        """Create a task that monitors the signal_received flag."""
        self.rgc.signal_monitor_task = asyncio.create_task(
            self._monitor_signal()
        )

    async def _monitor_signal(self):
        """Check if the signal_received was set."""
        while self.rgc.signal_received is None:
            try:
                await asyncio.sleep(0.5)
            except asyncio.CancelledError:
                break
        LOG.debug(
            "RunGovernor(id=0x%x, uid=%s) received signal %s.",
            id(self.rgc),
            self.rgc.uid,
            self.rgc.signal_received,
        )
        self.rgc.signal_received = None

    def _init_communication(self):
        self.rgc.major_domo_worker = MajorDomoWorker(
            broker_uri=self.rgc.broker_uri,
            service=self.rgc.uid,
        )
        LOG.debug(
            "RunGovernor(id=0x%x, uid=%s) started "
            "MajorDomoWorker(id=0x%x, uri=%s).",
            id(self.rgc),
            self.rgc.uid,
            id(self.rgc.major_domo_worker),
            self.rgc.broker_uri,
        )
        self.rgc.major_domo_client = MajorDomoClient(self.rgc.broker_uri)
        LOG.debug(
            "RunGovernor(id=0x%x, uid=%s) started "
            "MajorDomoClient(id=0x%x, uri=%s).",
            id(self.rgc),
            self.rgc.uid,
            id(self.rgc.major_domo_client),
            self.rgc.broker_uri,
        )



class SimControllerTerminationRequest:
    def __init__(
        self,
        sc_id,
        run_id,
        is_terminal,
        env_terminal,
        last_reward,
        additional_results,
    ):
        self.sc_id = sc_id
        self.additional_results = additional_results
        self.last_reward = last_reward
        self.env_terminal = env_terminal
        self.is_terminal = is_terminal
        self.run_id = run_id



from __future__ import annotations
from dataclasses import dataclass

# from typing import TYPE_CHECKING


@dataclass
class MuscleShutdownResponse:
    """Responds after a :class:`.Muscle` sent a
    :class:`.MuscleShutdownRequest` to the :class:`.Brain`.

    """



"""Majordomo Protocol Worker API, Python version
Implements the MDP/Worker spec at http:#rfc.zeromq.org/spec:7.
Author: Min RK <benjaminrk@gmail.com>
Based on Java example by Arkadiusz Orzechowski
"""

import time

import zmq
import zmq.asyncio

import palaestrai.core.MDP as MDP
from palaestrai.core.serialisation import deserialize, serialize

from . import LOG
from .zhelpers import pprint_message, ContextGuard


class MajorDomoWorker:
    """A task receiver with heartbeats

    The Major Domo Worker connects to a broker and listens to requests
    to designated service. Thiese it picks up, returns them to the calling
    loop, and also sends replies. The worker has a separate heartbeat to find
    out whether the connection to the broker has become stale.

    This is the Major Domo Protocol Worker API, Python version.
    Implements the MDP/Worker spec at http:#rfc.zeromq.org/spec:7.
    """

    HEARTBEAT_LIVENESS = 3  # 3-5 is reasonable
    DEFAULT_HEARTBEAT_DELAY = 2500  # Heartbeat delay, msecs
    reconnect = 2500  # Reconnect delay, msecs
    timeout = 2500  # poller timeout

    def __init__(self, broker_uri: str, service):
        self._socket = None
        self._broker_uri = broker_uri
        if isinstance(service, str):
            self._service = bytes(service, "ascii")
        elif isinstance(service, bytes):
            self._service = service
        else:
            self._service = bytes(str(service), "ascii")
        self._reply_to = None
        self._expect_reply = False
        self._poller = zmq.asyncio.Poller()
        self._liveness = 0
        self._heartbeat_at = 0
        self.heartbeat_delay = MajorDomoWorker.DEFAULT_HEARTBEAT_DELAY

    async def _connect(self):
        """Connect or reconnect to broker"""
        if self._socket:
            self._poller.unregister(self._socket)
            self._socket.close()
        self._socket = zmq.asyncio.Socket(
            context=ContextGuard.asyncio_context(), socket_type=zmq.DEALER
        )
        self._socket.linger = 0
        self._socket.connect(self._broker_uri)
        self._poller.register(self._socket, zmq.POLLIN)
        LOG.debug(
            "MajorDomoWorker(id=0x%x, service=%s, broker_uri=%s) "
            "connecting to MajorDomoBroker",
            id(self),
            self._service,
            self._broker_uri,
        )

        # Register service with broker
        await self._send(MDP.W_READY, self._service, [])

        # If liveness hits zero, queue is considered disconnected
        self._liveness = self.HEARTBEAT_LIVENESS
        self._heartbeat_at = time.time() + 1e-3 * self.heartbeat_delay

    async def _send(self, command, option=None, msg=None):
        """Send message to broker.
        If no msg is provided, creates one internally
        """

        if not self._socket:
            await self._connect()

        if msg is None:
            msg = []
        elif not isinstance(msg, list):
            msg = [msg]
        if option:
            msg = [option] + msg

        LOG.debug(
            "MajorDomoWorker(id=0x%x, service=%s, uri=%s) "
            "sending to broker: %s",
            id(self),
            self._service,
            self._broker_uri,
            msg,
        )
        msg = [b"", MDP.W_WORKER, command] + msg
        await self._socket.send_multipart(msg)

    async def transceive(self, reply=None, skip_recv=False):
        """Send and receive main method of the major domo worker

        This method does it both: First, it sends a reply to the last message
        that was recived --- the major domo worker keeps track of that ---,
        then it waits for the next message. This is the normal operation.

        There can be exceptions, which are only sensible for the first or
        the last message. One can skip sending a reply, which is sensible for
        the very first message (`reply=None`), or one can skip the receiving
        part (`skip_recv=True`), which is sensible when the parent object of
        the worker shuts down and sends its last ACK.

        :param reply: The reply to send to the last message received.
        :param skip_recv: If `True`, the worker will only send and skip the
            receiving part.
        :return: A message body in serialized form
        """
        LOG.debug(
            "MajorDomoWorker(id=0x%x, broker_uri=%s, service=%s) "
            "in recv(reply=%s, skip_recv=%s), starting to poll",
            id(self),
            self._broker_uri,
            self._service,
            reply,
            skip_recv,
        )
        if reply is None and self._expect_reply:
            LOG.error(
                "MajorDomoWorker(id=%s, broker_uri=%s, service=%s) "
                "is expected to send a reply, but has none on recv; cowardly"
                "refusing to continue",
                id(self),
                self._broker_uri,
                self._service,
            )
            return

        if reply:
            assert not isinstance(reply, bytes)
            reply = serialize(reply)
            if not isinstance(reply, list):
                reply = [reply]
            reply = [self._reply_to, b""] + reply
            await self._send(MDP.W_REPLY, msg=reply)
        self._expect_reply = True  # We may recv only once without reply.

        if skip_recv:
            return
        while True:
            # Poll socket for a reply, with timeout
            try:
                LOG.debug(
                    "MajorDomoWorker(id=0x%x, service=%s, broker_uri=%s) "
                    "waiting for messages",
                    id(self),
                    self._service,
                    self._broker_uri,
                )
                items = await self._poller.poll(self.timeout)
            except KeyboardInterrupt:
                break  # Interrupted

            if items:
                msg = await self._socket.recv_multipart()
                LOG.debug(
                    "MajorDomoWorker(id=0x%x, service=%s, broker_uri=%s) "
                    "received message from broker: %s",
                    id(self),
                    self._service,
                    self._broker_uri,
                    pprint_message(msg),
                )

                self._liveness = self.HEARTBEAT_LIVENESS
                # Don't try to handle errors, just assert noisily
                if len(msg) < 3:
                    LOG.error(
                        "MajorDomoWorker(id=0x%x, service=%s, "
                        "broker_uri=%s) "
                        "of length = %d, should be >= 3; ignoring",
                        id(self),
                        self._service,
                        self._broker_uri,
                        len(msg),
                    )
                    continue
                empty = msg.pop(0)
                if empty != b"":
                    LOG.error(
                        "MajorDomoWorker(id=0x%x, service=%s, "
                        "broker_uri=%s) received message that "
                        "violates the MDP: empty != ''",
                        id(self),
                        self._service,
                        self._broker_uri,
                    )
                    continue
                header = msg.pop(0)
                if header != MDP.W_WORKER:
                    LOG.error(
                        "MajorDomoWorker(id=0x%x, service=%s, "
                        "broker_uri=%s) received header '%s', but expected"
                        " '%s'; ignoring message",
                        id(self),
                        self._service,
                        self._broker_uri,
                        header,
                        MDP.W_WORKER,
                    )
                    continue

                command = msg.pop(0)
                if command == MDP.W_REQUEST:
                    # We should pop and save as many addresses as there are
                    # up to a null part, but for now, just save one...
                    self._reply_to = msg.pop(0)
                    # pop empty
                    empty = msg.pop(0)
                    assert empty == b""

                    return deserialize(msg)  # We have a request to process

                elif command == MDP.W_HEARTBEAT:
                    # Do nothing for heartbeats
                    pass
                elif command == MDP.W_DISCONNECT:
                    await self._connect()
                else:
                    LOG.error(
                        "MajorDomoWorker(id=0x%x, service=%s, "
                        "broker_uri=%s) "
                        "received invalid input message: %s",
                        id(self),
                        self._service,
                        self._broker_uri,
                        pprint_message(msg),
                    )
            else:
                self._liveness -= 1
                if self._liveness <= 0:
                    if self._socket:  # Reconnect: sleep a bit
                        LOG.warning(
                            "MajorDomoWorker(id=0x%x, service=%s, uri=%s) "
                            "disconnected from broker, reconnecting",
                            id(self),
                            self._service,
                            self._broker_uri,
                        )
                        try:
                            time.sleep(1e-3 * self.reconnect)
                        except KeyboardInterrupt:
                            break
                    await self._connect()

            # Send HEARTBEAT if it's time
            if time.time() > self._heartbeat_at:
                await self._send(MDP.W_HEARTBEAT)
                self._heartbeat_at = time.time() + 1e-3 * self.heartbeat_delay
        LOG.warning("Worker received interrupt, committing seppuku")
        return None

    def disconnect(self):
        """Disconnects the worker from the broker."""
        self._send(MDP.W_DISCONNECT)



import re
import ast
import numpy as np
from .space import Space


class MultiDiscrete(Space):
    """A series of discrete action spaces

    The multi-discrete action space consists of a series of discrete action
    spaces with different number of actions in each. It is useful to represent
    game controllers or keyboards where each key can be represented as a
    discrete action space. It is parametrized by passing an array of positive
    integers specifying number of actions for each discrete action space.

    For example:

        MultiDiscrete([ 5, 2, 2 ])
    """

    _MULTI_DISCRETE_RE = re.compile(r"MultiDiscrete\(([^)]+)\)")

    def __init__(self, nvec):
        """Constructs a new multi-discrete action space

        :param list[int] nvec: Vector of counts of each categorical variable
        """

        assert (np.array(nvec) > 0).all(), "nvec (counts) have to be positive"
        self.nvec = np.asarray(nvec, dtype=np.int64)
        super(MultiDiscrete, self).__init__(self.nvec.shape, np.int64)

    def sample(self):
        return (
            self.np_random.random_sample(self.nvec.shape) * self.nvec
        ).astype(self.dtype)

    def contains(self, x):
        if isinstance(x, list):
            x = np.array(x)  # Promote list to array for contains check
        # if nvec is uint32 and space dtype is uint32,
        # then 0 <= x < self.nvec guarantees that x is within correct bounds
        # for space dtype (even though x does not have to be unsigned)
        return (0 <= x).all() and (x < self.nvec).all()

    def to_string(self):
        return "MultiDiscrete(%s)" % np.array2string(self.nvec, separator=", ")

    @classmethod
    def from_string(cls, s):
        complete_match = MultiDiscrete._MULTI_DISCRETE_RE.match(s)
        if not complete_match:
            raise RuntimeError(
                "String '%s' does not match '%s'"
                % (s, MultiDiscrete._MULTI_DISCRETE_RE)
            )

        inner_str = complete_match[1]
        nvec = np.array(ast.literal_eval(inner_str))
        return MultiDiscrete(nvec)

    def __repr__(self):
        return self.to_string()

    def __eq__(self, other):
        return isinstance(other, MultiDiscrete) and np.all(
            self.nvec == other.nvec
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import Union


@dataclass
class ExperimentRunStartResponse:
    """Response to a :class:`Executor` if an experiment run was started.

    Parameters
    ----------
    sender_run_governor_id: str
        ID of the sending run governor.
    receiver_executor_id: str
        ID of the receiving executor.
    experiment_run_id: str
        ID of the started experiment run.
    successful: bool
        True, if the start was successful. False otherwise.
    error: Exception
        The error (message) if successful is False.

    """

    sender_run_governor_id: str
    receiver_executor_id: str
    experiment_run_id: str
    successful: bool
    error: Union[Exception, None]

    @property
    def sender(self):
        """ID of the sending run governor."""
        return self.sender_run_governor_id

    @property
    def receiver(self):
        """ID of the receiving executor."""
        return self.receiver_executor_id



import asyncio
import itertools
from typing import TYPE_CHECKING

import aiomultiprocess
from palaestrai.core.protocol import ExperimentRunShutdownRequest
from palaestrai.core.protocol import ExperimentRunStartRequest
from palaestrai.core.protocol import (
    SimulationControllerTerminationRequest as SimCtrlTermReq,
)
from palaestrai.util.exception import (
    DeadChildrenRisingAsZombiesError,
    InvalidRequestError,
    RequestIsNoneError,
    SignalInterruptError,
    TasksNotFinishedError,
)

from . import LOG, RunGovernorState

if TYPE_CHECKING:
    from palaestrai.experiment import RunGovernor


class RGSTransceiving(RunGovernorState):
    """Represent the TRANSCEIVING state of the run governor.

    This state is usually entered after INITIALIZING succeeds. Other
    states that may lead to this state are STARTING_RUN and
    HANDLING_SIM_CONTROLLER_TERMINATION.

    During this state, the run governor waits for incoming requests and
    sends replies to these requests. Depending on the request received,
    the run governor will transition to the next state handling this
    very request.

    Possible next states are
    * :class:`.RGSStarting on ExperimentStartRequest
    * :class:`.RGSHandlingSimControllerTermination on
       SimulationControllerTerminationRequest
    * :class:`.RGSStoppingSimulation on ExperimentShutdownRequest
    * :class:`.RGSErrorHandlingStarting in the error case.

    Parameters
    ----------
    rgc: :class:`.RunGovernor`
        The run governor instance that provides the context for this
        state.

    """

    def __init__(self, ctx: "RunGovernor"):
        super().__init__(ctx, "TRANSCEIVING")

    async def run(self):
        self._create_transceive_task()
        tasks_done, pending, all_processes = await self._wait_for_tasks()

        if not self._tasks_are_done(tasks_done):
            return

        if self._has_dead_children(all_processes):
            return

        await self._get_request(tasks_done)

        for task in pending:
            if task != self.rgc.signal_monitor_task:
                task.cancel()

    def next_state(self):
        from . import (
            RGSErrorHandlingTransceiving,
            RGSHandlingSimControllerTermination,
            RGSStartingSetup,
            RGSStoppingSimulation,
        )

        dispatcher = {
            ExperimentRunStartRequest: RGSStartingSetup,
            SimCtrlTermReq: RGSHandlingSimControllerTermination,
            ExperimentRunShutdownRequest: RGSStoppingSimulation,
        }
        try:
            state = dispatcher[self.rgc.last_request[0].__class__]
        except KeyError:
            self.add_error(
                InvalidRequestError(
                    [req for req in dispatcher],
                    self.rgc.last_request[0].__class__,
                )
            )
        except IndexError:
            # Is raised on signal interrupt because the last request
            # is none.
            pass

        if len(self.rgc.errors) > 0:
            state = RGSErrorHandlingTransceiving

        self.rgc.state = state(self.rgc)

    def _create_transceive_task(self):
        try:
            reply = self.rgc.next_reply.pop()
        except IndexError:
            reply = None

        self.rgc.transceive_task = asyncio.create_task(
            self.rgc.major_domo_worker.transceive(reply)
        )

    async def _wait_for_tasks(self):
        all_processes = [
            proc
            for proc in itertools.chain(
                self.rgc.sim_controllers.values(),
                self.rgc.env_conductors.values(),
                self.rgc.agent_conductors.values(),
            )
        ]

        LOG.debug(
            "RunGovernor(id=0x%x, uid=%s) "
            "in transceive and checking tasks (%s)",
            id(self.rgc),
            self.rgc.uid,
            [(proc.name, proc.is_alive()) for proc in all_processes]
            + [self.rgc.signal_monitor_task, self.rgc.transceive_task],
        )

        tasks_done, pending = await asyncio.wait(
            [
                asyncio.create_task(self._join_process(proc))
                for proc in all_processes
            ]
            + [self.rgc.signal_monitor_task, self.rgc.transceive_task],
            return_when=asyncio.FIRST_COMPLETED,
        )
        return tasks_done, pending, all_processes

    def _tasks_are_done(self, tasks_done) -> bool:
        if not tasks_done:
            # This shouldn't happen, but you never know
            self.add_error(TasksNotFinishedError())
            return False

        if self.rgc.signal_monitor_task in tasks_done:
            self.add_error(SignalInterruptError())
            return False

        return True

    def _has_dead_children(self, all_processes) -> bool:
        dead_children = [proc for proc in all_processes if not proc.is_alive()]

        if dead_children:
            self.add_error(DeadChildrenRisingAsZombiesError())
            return True
        return False

    async def _get_request(self, tasks_done):

        if self.rgc.transceive_task not in tasks_done:
            await self.rgc.transceive_task

        self.rgc.last_request.append(self.rgc.transceive_task.result())
        self.rgc.transceive_task = None
        if self.rgc.last_request[0] is None:
            LOG.warning(
                "RunGovernor(id=0x%x, uid=%s) "
                "received None request; ignoring",
                id(self.rgc),
                self.rgc.uid,
            )
            self.add_error(RequestIsNoneError())
            return False
        else:
            LOG.info(
                "RunGovernor(id=0x%x, uid=%s) received %s.",
                id(self.rgc),
                self.rgc.uid,
                self.rgc.last_request[0],
            )
        return True

    async def _join_process(self, process: aiomultiprocess.Process):
        LOG.debug(
            "RunGovernor(id=0x%x, uid=%s) "
            "monitoring Process(name=%s, is_alive=%s)",
            id(self.rgc),
            self.rgc.uid,
            process.name,
            process.is_alive(),
        )
        while process.exitcode is None:
            await asyncio.sleep(5)
            LOG.debug(
                "RunGovernor(id=0x%x, uid=%s) "
                "joined Process(name=%s, is_alive=%s, exitcode=%s, PID=%s)",
                id(self.rgc),
                self.rgc.uid,
                process.name,
                process.is_alive(),
                process.exitcode,
                process.pid,
            )



import unittest

import numpy as np
from palaestrai.agent import ActuatorInformation
from palaestrai.types import Box, Discrete, MultiBinary, MultiDiscrete, Tuple
from palaestrai.util.exception import OutOfActionSpaceError


class TestActuatorInformation(unittest.TestCase):
    def setUp(self):
        self.test_box1 = Box(0.0, 10.0, shape=(1,), dtype=np.float32)
        self.test_box2 = Box(
            low=np.array([0.0, -10.0]),
            high=np.array([1.0, 0.0]),
            dtype=np.float32,
        )
        self.test_box3 = Box(
            low=-1.0, high=2.0, shape=(3, 4), dtype=np.float32
        )
        self.test_discrete = Discrete(10)
        self.test_multidiscrete = MultiDiscrete([5, 4, 2])
        self.test_multibinary = MultiBinary(3)

    def test_call_actuator(self):
        act = ActuatorInformation(
            setpoint=None, action_space=self.test_box1, actuator_id="Test"
        )
        act(5.0)

        val = act.setpoint
        # NOTE: Is this the intended behavior?
        # NOTE: Shouldn't it be np.float32?
        self.assertIsInstance(val, float)
        self.assertEqual(val, 5.0)

    def test_set_one_dim_box(self):
        act = ActuatorInformation(
            setpoint=None, action_space=self.test_box1, actuator_id="Test"
        )

        # Not contained within the space
        with self.assertRaises(OutOfActionSpaceError):
            act.setpoint = 15.0
        self.assertIsNone(act.setpoint)

        # Wrong dimensions
        with self.assertRaises(OutOfActionSpaceError):
            act.setpoint = [5, 5]
        self.assertIsNone(act.setpoint)

        act.setpoint = 5.0
        self.assertIsInstance(act.setpoint, float)
        self.assertEqual(act.setpoint, 5.0)

    def test_set_two_dim_box(self):
        act = ActuatorInformation(
            setpoint=None, action_space=self.test_box2, actuator_id="Test"
        )

        # Wrong dimensions
        with self.assertRaises(OutOfActionSpaceError):
            act.setpoint = 0.5
        self.assertIsNone(act.setpoint)

        # Not within the space
        with self.assertRaises(OutOfActionSpaceError):
            act.setpoint = [-1, -5]
        self.assertIsNone(act.setpoint)

        act.setpoint = [0, -10]
        self.assertIsInstance(act.setpoint, list)

    def test_set_multi_dim_box(self):
        act = ActuatorInformation(
            setpoint=None, action_space=self.test_box3, actuator_id="Test"
        )

        # Wrong dimensions
        with self.assertRaises(OutOfActionSpaceError):
            act.setpoint = [0.0]
        self.assertIsNone(act.setpoint)

        # Not within the space
        with self.assertRaises(OutOfActionSpaceError):
            act.setpoint = [
                [0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0],
                [-2.0, 0.0, 0.0, 0.0],
            ]
        self.assertIsNone(act.setpoint)

        act.setpoint = [
            [0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0],
        ]
        self.assertIsInstance(act.setpoint, list)
        for sublist in act.setpoint:
            self.assertIsInstance(sublist, list)
            for val in sublist:
                self.assertIsInstance(val, float)

    def test_set_discrete(self):
        act = ActuatorInformation(
            setpoint=None, action_space=self.test_discrete, actuator_id="Test"
        )

        # Wrong dimensions
        with self.assertRaises(OutOfActionSpaceError):
            act.setpoint = [5, 5]

        # Not within the space
        with self.assertRaises(OutOfActionSpaceError):
            act.setpoint = -1

        # Wrong dtype
        with self.assertRaises(OutOfActionSpaceError):
            act.setpoint = 5.0

        act.setpoint = 5

    def test_set_multi_discrete(self):
        act = ActuatorInformation(None, self.test_multidiscrete, "Test")

        # Wrong dimensions
        with self.assertRaises(OutOfActionSpaceError):
            act.setpoint = [1, 1]

        # Not within the space
        with self.assertRaises(OutOfActionSpaceError):
            act.setpoint = [5, 4, 3]

        # Wrong dtype (should probably fail)
        # with self.assertRaises(OutOfActionSpaceError):
        act.setpoint = [3, 3, 1.0]

        act.setpoint = [1, 1, 1]
        self.assertIsInstance(act.setpoint, list)
        for val in act.setpoint:
            self.assertIsInstance(val, int)

    def test_set_multi_binary(self):
        act = ActuatorInformation(None, self.test_multibinary, "Test")

        with self.assertRaises(OutOfActionSpaceError):
            act.setpoint = [1, 1]
        with self.assertRaises(OutOfActionSpaceError):
            act.setpoint = [1, 1, 0, 0]

        # Not within the space
        with self.assertRaises(OutOfActionSpaceError):
            act.setpoint = [0, 1, 2]

        with self.assertRaises(OutOfActionSpaceError):
            act.setpoint = [0, 0.0, 1.0]

        act.setpoint = [0, 1, 0]

    def test_set_tuple(self):
        act = ActuatorInformation(
            None,
            Tuple(self.test_box2, self.test_discrete, self.test_multibinary),
            "Test",
        )

        # Wrong dimensions
        with self.assertRaises(OutOfActionSpaceError):
            act.setpoint = [-10]

        # Not within the space
        with self.assertRaises(OutOfActionSpaceError):
            act.setpoint = [[0, -5], 5, [0, 1, 0], [1.0]]

        act.setpoint = [[0, -5], 5, [0, 1, 0]]


if __name__ == "__main__":
    unittest.main()



import logging
import random
import sys
import unittest
from unittest import IsolatedAsyncioTestCase
from unittest.mock import AsyncMock, MagicMock, patch, call

from palaestrai.agent import RewardInformation
from palaestrai.agent.actuator_information import ActuatorInformation
from palaestrai.agent.sensor_information import SensorInformation
from palaestrai.core.protocol import (
    AgentSetupRequest,
    EnvironmentSetupRequest,
    EnvironmentStartRequest,
    EnvironmentStartResponse,
    EnvironmentUpdateResponse,
    SimulationControllerTerminationResponse,
)
from palaestrai.core.protocol.agent_update_rsp import AgentUpdateResponse
from palaestrai.core.serialisation import serialize
from palaestrai.simulation import VanillaSimControllerTerminationCondition
from palaestrai.simulation.vanilla_sim_controller import VanillaSimController
from palaestrai.types import Box, Discrete, Space


# from uuid import uuid4


def handle_client_msg(empf, message):

    if isinstance(message, EnvironmentSetupRequest):
        msg = serialize(
            EnvironmentUpdateResponse(
                "0",
                "0",
                "0",
                "0",
                [SensorInformation(1, Discrete(1), 0)],
                0.0,
                random.choice([True, False]),
            )
        )
        # print("returning msg")
        return msg

    elif isinstance(message, EnvironmentStartRequest):
        return serialize(
            EnvironmentStartResponse(
                0,
                0,
                [SensorInformation(1, Discrete(1), 0)],
                [
                    ActuatorInformation(
                        Discrete(1), Space.from_string("Discrete(2)"), 0
                    )
                ],
            )
        )
    elif isinstance(message, AgentSetupRequest):
        return serialize(
            AgentUpdateResponse(
                "0",
                "0",
                "0",
                [SensorInformation(1, Discrete(1), 0)],
                [
                    ActuatorInformation(
                        Discrete(1), Space.from_string("Discrete(2)"), 0
                    )
                ],
                [0.0],
            )
        )


class TestVanillaSimController(IsolatedAsyncioTestCase):
    def setUp(self):
        agent_dic = [
            {
                "name": "defender",
                "brain": {"name": "", "params": dict()},
                "muscle": {"name": "", "params": dict()},
                "objective": {"name": "", "params": dict()},
                "sensors": list(),
                "actuators": list(),
            },
        ]

        # uuid = uuid4()
        self.patch_client = (
            "palaestrai.simulation.simulation_controller.MajorDomoClient"
        )
        self.patch_worker = (
            "palaestrai.simulation.simulation_controller.MajorDomoWorker"
        )
        with patch(self.patch_worker, autospec=True):
            with patch(self.patch_client, autospec=True) as mock_client:
                mock_client.send = MagicMock(
                    AsyncMock(side_effect=handle_client_msg)
                )
                self.vsc = VanillaSimController(
                    "test",
                    "test2",
                    [1, 2],
                    [3],
                    agent_dic,
                    [
                        {
                            "name": (
                                "palaestrai.simulation:"
                                "VanillaSimControllerTerminationCondition"
                            ),
                            "params": {},
                        }
                    ],
                    mode="train",
                )

    async def test_term_condition_true(self):
        self.vsc.conductor_shutdown = AsyncMock(return_value=None)
        self.vsc.agent_shutdown = AsyncMock(return_value=None)
        self.vsc.env_shutdown = AsyncMock(return_value=None)

        msg = SimulationControllerTerminationResponse(
            "0", "1", "2", False, False
        )
        self.vsc.run_gov_client.send = AsyncMock(return_value=msg)

        await self.vsc.simulation_shutdown(True, [1], None)
        self.vsc.agent_shutdown.assert_called_once()
        self.vsc.env_shutdown.assert_called_once()

    async def test_term_condition_no_complete_shutdown(self):
        self.vsc.conductor_shutdown = MagicMock(return_value=None)
        self.vsc.agent_shutdown = AsyncMock(return_value=None)
        self.vsc.env_shutdown = AsyncMock(return_value=None)
        msg = SimulationControllerTerminationResponse(
            "0", "1", "2", False, False
        )
        self.vsc.run_gov_client.send = AsyncMock(return_value=msg)

        await self.vsc.simulation_shutdown(True, [1], None)
        self.vsc.conductor_shutdown.assert_not_called()

    def test_env_update(self):
        rsp_list = []
        sensor_list = []
        reward_list = []
        for i in range(4):
            sen = [SensorInformation(1, Discrete(1), i)]
            sensor_list.extend(sen)
            reward = RewardInformation(i * 2, Discrete(1), "Test")
            reward_list.append(reward)
            msg = EnvironmentUpdateResponse(
                sender_environment_id="0",
                receiver_simulation_controller_id="0",
                environment_conductor_id="0",
                experiment_run_id=str(i),
                sensors=sen,
                reward=[reward],
                is_terminal=True,
            )
            rsp_list.append(msg)

        sensors, rewards, termination = self.vsc.env_update(rsp_list)

        self.assertTrue(termination)
        self.assertListEqual(rewards, reward_list)
        self.assertListEqual(sensors, sensor_list)

    def test_agent_update(self):
        rsp_list = []
        actuator_list = []
        for i in range(4):
            actuator = [ActuatorInformation(i, Discrete(1), i)]
            actuator_list.extend(actuator)
            msg = AgentUpdateResponse(
                sender_agent_id="a",
                receiver_simulation_controller_id="s",
                actuator_information=actuator,
                sensor_information=[],
                experiment_run_id=str(i),
            )
            rsp_list.append(msg)

        actuators = self.vsc.agent_update(rsp_list)

        self.assertListEqual(actuators, actuator_list)

    async def test_simulation_stop_at_termination(self):
        reward = [RewardInformation(0.0, Discrete(1), "Test")]
        self.vsc.termination_condition = (
            VanillaSimControllerTerminationCondition()
        )
        self.vsc.env_ids = [0]
        self.vsc.get_env_update = AsyncMock(
            return_value=EnvironmentUpdateResponse(
                sender_environment_id="0",
                receiver_simulation_controller_id="0",
                environment_conductor_id="0",
                experiment_run_id="0",
                sensors=[SensorInformation(1, Discrete(1), 0)],
                reward=reward,
                is_terminal=True,
            )
        )
        self.vsc.get_agent_update = AsyncMock(
            return_value=AgentUpdateResponse(
                sender_agent_id="a",
                receiver_simulation_controller_id="s",
                actuator_information=[],
                sensor_information=[],
                experiment_run_id=str("e"),
            )
        )
        self.vsc.simulation_shutdown = MagicMock(AsyncMock(return_value=None))
        self.vsc.continue_simulation = AsyncMock(return_value=False)

        await self.vsc.simulation()
        self.vsc.continue_simulation.assert_called_once_with(
            True, reward, None
        )

    async def test_continue_simulation_no_termination(self):
        self.vsc.ex_termination = False
        result = await self.vsc.continue_simulation(False, 0, dict())

        self.assertTrue(result)

    async def test_continue_simulation_ex_termination(self):
        self.vsc.ex_termination = True
        self.vsc.stop_simulation = AsyncMock()

        result = await self.vsc.continue_simulation(False, 0, dict())

        self.assertFalse(result)

    async def test_continue_simulation_restart(self):
        self.vsc.ex_termination = False
        self.vsc.send_termination_request = AsyncMock(
            return_value=SimulationControllerTerminationResponse(
                "0", "1", "2", True, False
            )
        )
        self.vsc.stop_simulation = AsyncMock()

        result = await self.vsc.continue_simulation(True, 0, dict())

        self.assertTrue(result)

    async def test_continue_simulation_termination(self):
        self.vsc.ex_termination = False
        self.vsc.send_termination_request = AsyncMock(
            return_value=SimulationControllerTerminationResponse(
                "0", "1", "2", True, True
            )
        )
        self.vsc.stop_simulation = AsyncMock()

        result = await self.vsc.continue_simulation(True, 0, dict())

        self.assertFalse(result)

    def test_access_list(self):
        wanted_sensors = ["test-1", "test-2", "test-3"]
        wanted_actuators = ["test-4", "test-5", "test-6"]

        self.vsc.all_actuators = []
        self.vsc.all_actuators = []
        self.assertEqual(
            self.vsc.access_list(wanted_sensors, wanted_actuators), ([], [])
        )

        self.vsc.all_sensors = [
            SensorInformation(0, Discrete(1), "test-1"),
            SensorInformation(0, Discrete(1), "test-4"),
        ]
        self.vsc.all_actuators = []
        self.assertEqual(
            self.vsc.access_list(wanted_sensors, wanted_actuators),
            ([self.vsc.all_sensors[0]], []),
        )

        self.vsc.all_actuators = [
            ActuatorInformation(
                Box(0, 1, shape=(1,)), Discrete(1), "test-1337"
            ),
            ActuatorInformation(Box(0, 1, shape=(1,)), Discrete(1), "test-5"),
        ]
        self.assertEqual(
            self.vsc.access_list(wanted_sensors, wanted_actuators),
            ([self.vsc.all_sensors[0]], [self.vsc.all_actuators[1]]),
        )


if __name__ == "__main__":
    logging.basicConfig(stream=sys.stderr, level=logging.DEBUG)
    unittest.main()



# encoding: utf-8
"""
Helper module for example applications. Mimics ZeroMQ Guide's zhelpers.h.
"""

from __future__ import print_function

import binascii
import os
import socket
from random import randint

import zmq
import zmq.asyncio


def socket_set_hwm(socket, hwm=-1):
    """libzmq 2/3/4 compatible sethwm"""
    try:
        socket.sndhwm = socket.rcvhwm = hwm
    except AttributeError:
        socket.hwm = hwm


def pprint_message(msg):
    """Pretty-prints a message to a string.

    :param msg: A message of the palaestrai zmq core protocol
    :return: The pretty-printed message
    """

    def str_or_hex(p):
        try:
            return p.decode("ascii")
        except UnicodeDecodeError:
            return r"0x%s" % binascii.hexlify(p).decode("ascii")

    return "\n".join(
        ["[%03d] %s" % (len(p), str_or_hex(p)) for p in [x for x in msg]]
    )


def dump(msg_or_socket):
    """Receives all message parts from socket, printing each frame neatly"""
    if isinstance(msg_or_socket, zmq.Socket):
        # it's a socket, call on current message
        msg = msg_or_socket.recv_multipart()
    else:
        msg = msg_or_socket
    print("----------------------------------------")
    for part in msg:
        print("[%03d]" % len(part), end=" ")
        is_text = True
        try:
            print(part.decode("ascii"))
        except UnicodeDecodeError:
            print(r"0x%s" % (binascii.hexlify(part).decode("ascii")))


def set_id(zsocket):
    """Set simple random printable identity on socket"""
    identity = "%04x-%04x" % (randint(0, 0x10000), randint(0, 0x10000))
    zsocket.setsockopt_string(zmq.IDENTITY, identity)


def zpipe(ctx):
    """build inproc pipe for talking to threads
    mimic pipe used in czmq zthread_fork.
    Returns a pair of PAIRs connected via inproc
    """
    a = ctx.socket(zmq.PAIR)
    b = ctx.socket(zmq.PAIR)
    a.linger = b.linger = 0
    a.hwm = b.hwm = 1
    iface = "inproc://%s" % binascii.hexlify(os.urandom(8))
    a.bind(iface)
    b.connect(iface)
    return a, b


class _ContextGuard:
    """Guards ZMQ contexts: Creates exactly one per process

    This guard class makes sure that only ever one ZMQ Context is created per
    process. It automagically creates a new one when there's a need to (e.g.,
    after fork()), but returns the current one otherwise.
    """

    def __init__(self):
        self._id = None
        self._context = None
        self._asyncio_context = None

    @staticmethod
    def _generate_id():
        return "%s@%s" % (os.getpid(), socket.gethostname())

    def _update_contexts(self):
        fresh_id = self._generate_id()
        if not self._context or not self._id or self._id != fresh_id:
            self._id = fresh_id
            self._context = zmq.Context()
            self._asyncio_context = zmq.asyncio.Context()

    def context(self):
        self._update_contexts()
        return self._context

    def asyncio_context(self):
        self._update_contexts()
        return self._asyncio_context


ContextGuard = _ContextGuard()



import asyncio
import itertools
import os
import signal
from typing import TYPE_CHECKING

import aiomultiprocess

from . import LOG, RunGovernorState

if TYPE_CHECKING:
    from palaestrai.experiment import RunGovernor


class RGSFinalizing(RunGovernorState):
    """Represents the FINALIZING state.

    Possible next states are
    * :class:`.RGSDone`.
    """

    def __init__(self, rgc: "RunGovernor"):
        super().__init__(rgc, "FINALIZING")

    async def run(self):
        await self._shutdown_tasks()
        LOG.debug(
            "RunGovernor(id=0x%x, uid=%s) completed shutdown",
            id(self.rgc),
            self.rgc.uid,
        )

    def next_state(self):
        from . import RGSDone

        self.rgc.state = RGSDone(self.rgc)

    async def _shutdown_tasks(self):
        """Terminates all running subprocesses and waits for their end."""

        asyncio.get_running_loop().remove_signal_handler(signal.SIGTERM)
        LOG.debug(
            "RunGovernor(id=0x%x, uid=%s) terminating process group. "
            "If there are no leftovers, this RunGovernor leaves for "
            "good.",
            id(self.rgc),
            self.rgc.uid,
        )
        os.killpg(0, signal.SIGTERM)
        LOG.debug(
            "RunGovernor(id=0x%x, uid=%s) saw processes surviving. "
            "Let the reaping begin.",
            id(self.rgc),
            self.rgc.uid,
        )
        await asyncio.gather(
            *[
                RGSFinalizing._reap_process(p)
                for p in itertools.chain(
                    self.rgc.agent_conductors.values(),
                    self.rgc.env_conductors.values(),
                    self.rgc.sim_controllers.values(),
                )
            ]
        )

        self.rgc.agent_conductors = {}
        self.rgc.sim_controllers = {}
        self.rgc.env_conductors = {}

    @staticmethod
    async def _reap_process(process: aiomultiprocess.Process):
        """Terminates a process, joins and reaps it.

        :param process: The process we have to reap
        """
        if process.is_alive():
            process.terminate()
        try:
            await process.join(timeout=3)
        except asyncio.TimeoutError:
            if process.is_alive():
                process.kill()



"""This module contains the class :class:`EnvironmentConductor` that
controls the creation of new environments.

"""

import asyncio
import logging
import signal
from typing import List
from uuid import uuid4

import aiomultiprocess
import setproctitle

from ..core import BasicState, MajorDomoWorker, RuntimeConfig
from ..core.protocol import (
    EnvironmentSetupRequest,
    EnvironmentSetupResponse,
    ShutdownRequest,
    ShutdownResponse,
)
from ..util.dynaloader import load_with_params
from palaestrai.util import spawn_wrapper

LOG = logging.getLogger("palaestrai.environment.conductor")


class EnvironmentConductor:
    """The environment conductor creates new environment instances.

    There could be multiple simulation runs and each would need a
    separate environment. The environment conductor controls the
    creation of those new environment instances.

     Parameters
     ----------
     env_type : str
        Path to the used environment, it is specified in the
        experiment.yml
     env_params : dict
        Dictionary with parameters needed by the environment
     simulation_controller_connection : str
        URI used to connect to the simulation broker
     uid : uuid4
        Unique identifier

    """

    def __init__(self, env_cfg, reward_cfg, broker_uri, seed: int, uid=None):
        self.seed = seed
        self._uid = str(uid) if uid else "EnvironmentConductor-%s" % uuid4()
        self.env_cfg = env_cfg
        self.reward_cfg = reward_cfg
        # self.env_type = env_type
        # self.env_params = env_params
        self._broker_uri = broker_uri
        self._state = BasicState.PRISTINE

        self._tasks: List[aiomultiprocess.Process] = []
        self._worker = None

        LOG.debug(
            "EnvironmentConductor(id=0x%x, uid=%s) created.",
            id(self),
            self.uid,
        )

    @property
    def uid(self):
        return self._uid

    @property
    def worker(self):
        if not self._worker:
            self._worker = MajorDomoWorker(
                broker_uri=self._broker_uri,
                service=self.uid,
            )
        return self._worker

    def _handle_signal_interrupt(self, signum):
        """Handle interrupting signals by notifying of the state change."""
        LOG.info(
            "EnvironmentConductor(id=0x%x, uid=%s) interrupted by signal %s.",
            id(self),
            self.uid,
            signum,
        )
        self._state = {
            signal.SIGINT.value: BasicState.SIGINT,
            signal.SIGABRT.value: BasicState.SIGABRT,
            signal.SIGTERM.value: BasicState.SIGTERM,
        }[signum]

    def _init_signal_handler(self):
        """Sets handlers for interrupting signals in the event loop."""
        signals = {signal.SIGINT, signal.SIGABRT, signal.SIGTERM}
        LOG.debug(
            "EnvironmentConductor(id=0x%x, uid=%s) "
            "registering signal handlers for signals %s.",
            id(self),
            self.uid,
            signals,
        )
        loop = asyncio.get_running_loop()
        for signum in signals:
            loop.add_signal_handler(
                signum, self._handle_signal_interrupt, signum
            )

    async def _monitor_state(self):
        known_state = self._state
        while known_state.value == self._state.value:
            try:
                await asyncio.sleep(0.5)
            except asyncio.CancelledError:
                break
        LOG.debug(
            "EnvironmentConductor(id=0x%x, uid=%s) "
            "state changed from %s to %s.",
            id(self),
            self.uid,
            known_state,
            self._state,
        )

    def init_env(self, uid):
        """Initialize a new environment.

        Creates a new environment instance with its own UID.

        Parameters
        ----------
        uid : uuid4
            Unique identifier

        """
        reward = load_with_params(
            self.reward_cfg["name"],
            self.reward_cfg.get("params", dict()),
        )
        env = load_with_params(
            self.env_cfg["name"],
            {
                "connection": self._broker_uri,
                "uid": uid,
                "reward": reward,
                "seed": self.seed,
                "params": self.env_cfg.get("params", dict()),
            },
        )
        LOG.debug(
            "EnvironmentConductor(id=0x%x, uid=%s) "
            "loaded Environment %s(id=0x%x, uid=%s).",
            id(self),
            self.uid,
            env.__class__,
            id(env),
            env.uid,
        )
        try:
            env_process = aiomultiprocess.Process(
                name=f"Environment-{self.uid}",
                target=spawn_wrapper,
                args=(
                    f"Environment-{self.uid}",
                    RuntimeConfig().to_dict(),
                    env.run,
                ),
            )
            env_process.start()
            self._tasks.append(env_process)
        except Exception as e:
            LOG.critical(
                "EnvironmentConductor(id=0x%x, uid=%s) "
                "encountered a fatal error while executing "
                "Environment(id=0x%x, uid=%s): %s.",
                id(self),
                self.uid,
                id(env),
                env.uid,
                e,
            )
            raise e

    async def run(self):
        signal.signal(signal.SIGINT, signal.SIG_DFL)
        signal.signal(signal.SIGABRT, signal.SIG_DFL)
        signal.signal(signal.SIGTERM, signal.SIG_DFL)
        setproctitle.setproctitle(
            "palaestrAI[EnvironmentConductor-%s]" % self._uid[-6:]
        )

        self._init_signal_handler()
        self._state = BasicState.RUNNING
        request = None
        reply = None
        LOG.info(
            "EnvironmentConductor(id=0x%x, uid=%s) commencing run: "
            "creating better worlds.",
            id(self),
            self.uid,
        )

        state_monitor_task = asyncio.create_task(self._monitor_state())
        while self._state == BasicState.RUNNING:
            transceive_task = asyncio.create_task(
                self.worker.transceive(reply)
            )
            tasks_done, tasks_pending = await asyncio.wait(
                {state_monitor_task, transceive_task},
                return_when=asyncio.FIRST_COMPLETED,
            )
            if transceive_task in tasks_pending:
                continue
            request = transceive_task.result()
            if not request:
                continue

            LOG.info(
                "EnvironmentConductor(id=0x%x, uid=%s) "
                "received request: %s(%s).",
                id(self),
                self.uid,
                request.__class__,
                request.__dict__,
            )

            if isinstance(request, EnvironmentSetupRequest):
                LOG.debug(
                    "EnvironmentConductor(id=0x%x, uid=%s) "
                    "received EnvironmentSetupRequest(experiment_run_id=%s).",
                    id(self),
                    self.uid,
                    request.experiment_run_id,
                )
                self.init_env(request.environment_id)
                ssci = request.sender_simulation_controller_id
                reply = EnvironmentSetupResponse(
                    sender_environment_conductor=self.uid,
                    receiver_simulation_controller=ssci,
                    environment_id=request.environment_id,
                    experiment_run_id=request.experiment_run_id,
                    environment_type=self.env_cfg["name"],
                    environment_parameters=self.env_cfg.get("params", dict()),
                )
            if isinstance(request, ShutdownRequest):
                self._state = BasicState.STOPPING
                for task in self._tasks:
                    await task.join()

        reply = ShutdownResponse(
            request.experiment_run_id if request else None
        )
        await self.worker.transceive(reply, skip_recv=True)
        self._state = BasicState.FINISHED
        LOG.info(
            "EnvironmentConductor(id=0x%x, uid=%s) completed shutdown.",
            id(self),
            self.uid,
        )



"""
This module contains the class :class:`DummyEnvironment`. It could be
used in an experiment for reference purposes.
"""
import random
from typing import List

from palaestrai.agent import SensorInformation, ActuatorInformation
from palaestrai.types import Discrete
from .environment import Environment
from ..agent.reward_information import RewardInformation


class DummyEnvironment(Environment):
    """
    This class provides a dummy environment with a fixed number of sensors. The
    environment terminates after a fixed number of updates.

    Parameters
    ----------
    connection : broker_connection
        the URI which is used to connect to the simulation broker. It is used
        to communicate with the simulation controller.
    uid : uuid4
        a universal id for the environment
    seed : int
        Seed for recreation
    params : dict
        parameters which are necessary for the environment execution
        default to None
    """

    def calc_reward(
        self,
        sensors: List[SensorInformation],
        actuators: List[ActuatorInformation],
    ) -> List[RewardInformation]:
        return [RewardInformation(0, Discrete(2), "Reward")]

    def __init__(self, connection, uid, reward, seed: int, params=None):
        super().__init__(connection, uid, reward, seed, params)
        self.iter: int

    def start_environment(self):
        """
        This method is called when an `EnvironmentStartRequest` message is
        received. This dummy environment is represented by 10 sensors and
        10 actuators. The sensors are of the type `SensorInformation` and have
        a random value of either 0 or 1, an `observation_space` between 0 and 1
        and an integer number as id.
        The actuators are of the type `ActuatorInformation` and contain a
        setpoint of Discrete(1), an `action_space` of None and an integer
        number as id.

        Returns
        -------
        tuple :
            A list containing the `SensorInformation` for each of the 10
            sensors and a list containing the `ActuatorInformation` for each
            of the 10 actuators.
        """
        self.iter = 0
        sensors = []
        for num in range(10):
            sensors.append(self._create_sensor(num))
            actuator = ActuatorInformation(
                action_space=Discrete(2), setpoint=None, actuator_id=num
            )
            self.actuators.append(actuator)
        return sensors, self.actuators

    def update(self, actuators):
        """
        This method is called when an `EnvironmentUpdateRequest` message is
        received. While setpoints of the actuators manipulate an actual
        environment, in here those setpoints have no impact on the behavior of
        the dummy environment.
        The state of this dummy environment is represented via random values of
        the `SensorInformation` from the 10 sensors.
        In this dummy environment the reward for the state is a random value of
        either 0 or 1.
        The method returns a list of `SensorInformation`, the random reward and
        the boolean `is_terminal`. After 10 updates the `is_terminal` value is
        set to True which triggers the respective shutdown messages.

        Parameters
        ----------
        actuators : list[`ActuatorInformation`]
            A list of `ActuatorInformation` to interact with the environment.

        Returns
        -------
        tuple :
            A list of `SensorInformation` representing the 10 sensors, the
            random reward and boolean for `is_terminal`.

        """
        if self.iter < 10:
            self.iter += 1
            sensors = []
            for num in range(10):
                sensors.append(self._create_sensor(num))
            return sensors, False
        else:
            return list(), True

    def _create_sensor(self, sensor_id):
        return SensorInformation(
            sensor_value=random.randint(0, 1),
            observation_space=Discrete(2),
            sensor_id=sensor_id,
        )



import re
from .space import Space


class Tuple(Space):
    """A tuple (i.e., product) of simpler spaces

    Example usage:
    self.observation_space = spaces.Tuple(Discrete(2), Discrete(3))
    """

    _TUPLE_RE = re.compile(r"\A\s*?Tuple\((.+)\)\s*\Z")
    _INNER_PIECE_RE = re.compile(
        r"(?P<inner_rest>,\s*" r"(?P<piece>[A-Za-z]+\(.*\)))\s*\Z"
    )

    def __init__(self, *spaces):
        self.spaces = tuple(spaces)
        for space in spaces:
            assert isinstance(
                space, Space
            ), "Elements of the tuple must be instances of palaestrai.types.Space"
        super(Tuple, self).__init__(None, None)

    def seed(self, seed=None):
        [space.seed(seed) for space in self.spaces]

    def sample(self):
        return tuple([space.sample() for space in self.spaces])

    def contains(self, x):
        if isinstance(x, list):
            x = tuple(x)  # Promote list to tuple for contains check
        return (
            isinstance(x, tuple)
            and len(x) == len(self.spaces)
            and all(
                space.contains(part) for (space, part) in zip(self.spaces, x)
            )
        )

    def __repr__(self):
        return self.to_string()

    def to_string(self):
        return "Tuple(" + ", ".join([str(s) for s in self.spaces]) + ")"

    @classmethod
    def from_string(cls, s):
        complete_match = Tuple._TUPLE_RE.match(s)
        if not complete_match:
            raise RuntimeError(
                "String '%s' does not match '%s'" % (s, Tuple._TUPLE_RE)
            )
        inner_str = complete_match[1]

        spaces = []
        while len(inner_str) > 0:
            match = Tuple._INNER_PIECE_RE.search(inner_str)
            if match is None:
                try:
                    spaces.append(Space.from_string(inner_str))
                except:
                    pass  # We simply ignore garbage.
                break
            else:
                head, _, tail = inner_str.rpartition(match["inner_rest"])
                inner_str = head + tail
                spaces.append(Space.from_string(match["piece"]))
        spaces.reverse()

        return Tuple(*spaces)

    def __getitem__(self, index):
        return self.spaces[index]

    def __len__(self):
        return len(self.spaces)

    def __eq__(self, other):
        return isinstance(other, Tuple) and self.spaces == other.spaces



"""This module contains the debug script that can be used to start
the dummy experiment without the CLI.

"""
import os

from click.testing import CliRunner, Result

from palaestrai.cli.manager import cli

debug_script_path: str = os.path.abspath(
    f"{__file__}/../../fixtures/dummy_run.yml"
)

runtime_path: str = os.path.abspath(
    f"{__file__}/../../fixtures/palaestrai-runtime-debug.conf.yaml"
)


def test_system_experiment_start() -> None:
    runner: CliRunner = CliRunner()
    result: Result = runner.invoke(
        cli, ["-c", runtime_path, "-vv", "experiment-start", debug_script_path]
    )
    assert result.exit_code == 0


if __name__ == "__main__":
    test_system_experiment_start()



#!/usr/bin/env python3
"""Setup file for the ARL package."""
from setuptools import find_packages, setup


with open("VERSION") as freader:
    VERSION = freader.readline().strip()

with open("README.rst") as freader:
    README = freader.read()

install_requirements = [
    # CLI
    "click==7.1.2",
    "appdirs==1.4.4",
    # YAML
    "ruamel.yaml==0.17.4",
    # Process and IPC handling
    "aiomultiprocess==0.9.0",
    "setproctitle==1.2.2",
    "pyzmq==22.0.3",
    # Data handling and storage
    "alembic==1.5.8",
    "numpy==1.18.5",
    "pandas==1.2.4",
    "psycopg2-binary==2.8.6",
    "jsonpickle==2.0.0",
    "SQLalchemy < 1.4.0",
    "sqlalchemy_utils",
]

development_requirements = [
    # Tests
    "tox==3.23.0",
    "robotframework >= 4.0.0",
    "pytest==6.2.4",
    "pytest-asyncio",
    "pytest-cov",
    "coverage",
    "mypy",
    "black",
    "lxml",
]

extras = {"dev": development_requirements}

setup(
    name="palaestrai",
    version=VERSION,
    description="A Training Ground for Autonomous Agents",
    long_description=README,
    author="The ARL Developers",
    author_email="eric.veith@offis.de",
    python_requires=">=3.8.0",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    include_package_data=True,
    install_requires=install_requirements,
    extras_require=extras,
    license="LGPLv2",
    url="http://palaestr.ai/",
    entry_points="""
        [console_scripts]
        palaestrai=palaestrai.cli.manager:cli
        arl-apply-migrations=palaestrai.store.migrations.apply:main
    """,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: "
        "GNU Lesser General Public License v2 (LGPLv2)",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)



import unittest
from unittest.mock import AsyncMock, MagicMock, patch

from palaestrai.core import MajorDomoClient, MajorDomoWorker
from palaestrai.experiment import RunGovernor
from palaestrai.experiment.rungovernor_states import (
    RGSErrorHandlingInitializing,
    RGSInitializing,
    RGSTransceiving,
)


class TestInitializing(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.rungov = RunGovernor(None, None)
        self.rgs = RGSInitializing(self.rungov)
        self.rungov.state = self.rgs

    async def test_run(self):
        self.rgs._init_signal_handler = MagicMock()
        self.rgs._init_signal_monitor = MagicMock()
        self.rgs._init_communication = MagicMock()

        await self.rgs.run()

        self.rgs._init_signal_handler.assert_called_once()
        self.rgs._init_signal_monitor.assert_called_once()
        self.rgs._init_communication.assert_called_once()

    def test_next_state_success(self):

        self.rgs.next_state()

        self.assertIsInstance(self.rungov.state, RGSTransceiving)

    def test_next_state_failure(self):

        self.rgs.add_error(ValueError())
        self.rgs.next_state()

        self.assertIsInstance(self.rungov.state, RGSErrorHandlingInitializing)

    @patch(f"{RGSInitializing.__module__}.asyncio")
    def test_init_signal_handler(self, mock_asyncio):
        mock_loop = MagicMock()
        mock_loop.add_signal_handler = MagicMock()
        mock_asyncio.get_running_loop = MagicMock(return_value=mock_loop)

        self.rgs._init_signal_handler()

        self.assertEqual(3, mock_loop.add_signal_handler.call_count)

    def test_handle_signal_termination(self):
        signum = 0

        logger = f"{RunGovernor.__module__}".rsplit(".", 1)[0]

        with self.assertLogs(logger, level="INFO") as cm:
            self.rgs._handle_signal_termination(signum)

        self.assertEqual(signum, self.rungov.signal_received)
        self.assertIn("interrupted", cm.output[0])

    def test_handle_signal_interrupt(self):
        logger = f"{RunGovernor.__module__}".rsplit(".", 1)[0]

        with self.assertLogs(logger, level="DEBUG") as cm:
            self.rgs._handle_signal_interrupt()

        self.assertIn("SIGINT", cm.output[0])

    @patch(f"{RGSInitializing.__module__}.asyncio")
    def test_init_signal_monitor(self, mock_asyncio):
        mock_create = MagicMock(return_value=0)
        mock_asyncio.create_task = mock_create
        self.rgs._monitor_signal = MagicMock()

        self.rgs._init_signal_monitor()

        mock_create.assert_called_once()
        self.rgs._monitor_signal.assert_called_once()

    @patch(f"{RGSInitializing.__module__}.asyncio")
    async def test_monitor_signal(self, mock_asyncio):
        self.rungov.signal_received = 0

        mock_asyncio.sleep = AsyncMock()
        logger = f"{RunGovernor.__module__}".rsplit(".", 1)[0]
        with self.assertLogs(logger, level="DEBUG") as cm:
            await self.rgs._monitor_signal()

        self.assertIn("received signal", cm.output[0])
        self.assertIsNone(self.rungov.signal_received)

    def test_init_communication(self):

        self.rgs._init_communication()
        self.assertIsInstance(self.rungov.major_domo_worker, MajorDomoWorker)
        self.assertIsInstance(self.rungov.major_domo_client, MajorDomoClient)


if __name__ == "__main__":
    unittest.main()



import os
from pathlib import Path

from ruamel.yaml import YAML

yaml = YAML(typ="safe")
POSTGRES_ENVIRONMENT_VARIABLES = {
    "POSTGRES_HOST",
    "POSTGRES_USER",
    "POSTGRES_PASSWORD",
    "PGDB",
}


def prepare_for_store_test(source_path, target_path):
    conf = yaml.load(Path(source_path))
    rewrite_config = all(
        k in os.environ for k in POSTGRES_ENVIRONMENT_VARIABLES
    )
    if rewrite_config:
        conf["store_uri"] = "postgres://%s:%s@%s/%s" % (
            os.environ["POSTGRES_USER"],
            os.environ["POSTGRES_PASSWORD"],
            os.environ["POSTGRES_HOST"],
            os.environ["PGDB"],
        )
        conf["logging"]["loggers"]["palaestrai.store"]["level"] = "DEBUG"
    with open(target_path, "w") as fp:
        yaml.dump(conf, fp)


def prepare_for_sqlite_store_test(source_path, target_path, tmpdir):
    conf = yaml.load(Path(source_path))
    print(source_path)
    print(conf)
    assert conf
    conf["store_uri"] = "sqlite:///%s/palaestrai.db" % tmpdir
    conf["logging"]["loggers"]["palaestrai.store"]["level"] = "DEBUG"
    with open(target_path, "w") as fp:
        yaml.dump(conf, fp)



import unittest
from palaestrai.environment.dummy_environment import DummyEnvironment
from palaestrai.environment.dummy_reward import DummyReward


class DummyEnvironmentTest(unittest.TestCase):
    def setUp(self) -> None:
        self.env = DummyEnvironment(
            connection="test_conn",
            uid=0,
            reward=DummyReward(),
            seed=123,
            params={"test": 0},
        )
        self.env.start_environment()

    def test_sensors_actuators(self):
        self.assertTrue(
            (len(self.env.actuators)) != 0,
            "expected at least one actuator",
        )

    def test_iteration(self):
        for _ in range(10):
            self.assertNotEqual(
                (list(), True),
                self.env.update(self.env.actuators),
                "Environment terminated unexpectedly",
            )

        self.assertEqual(
            (list(), True),
            self.env.update(self.env.actuators),
            "Environment did not terminate on time",
        )


if __name__ == "__main__":
    unittest.main()



class MuscleUpdateResponse:
    def __init__(self, is_updated, updates):
        self._is_updated = is_updated
        self._updates = updates

    @property
    def is_updated(self):
        return self._is_updated

    @is_updated.setter
    def is_updated(self, value):
        self._is_updated = value

    @property
    def updates(self):
        return self._updates

    @updates.setter
    def updates(self, value):
        self._updates = value



import dataclasses
from typing import Union


@dataclasses.dataclass
class SimulationControllerTerminationResponse:
    """Acknowledges the termination of a :py:class:`SimulationController`.

    * Sender: :py:class:`RunGovernor`
    * Receiver: :py:class:`SimulationController`

    :param sender_run_governor_id: Opaque ID of the sending
        :py:class:`RunGovernor` instance
    :param receiver_simulation_controller_id: Opaque ID of the receiving
        :py:class:`SimulationController` instance
    :param complete_shutdown: If `True`, the message indicates that the
        :py:class:`RunGovernor` is now shutting down the whole run.
    """

    sender_run_governor_id: str
    receiver_simulation_controller_id: str
    experiment_run_id: Union[str, None]
    restart: bool
    complete_shutdown: bool

    @property
    def sender(self):
        return self.sender_run_governor_id

    @property
    def receiver(self):
        return self.receiver_simulation_controller_id



import asyncio
import sys

import click
import ruamel.yaml
from sqlalchemy.exc import ArgumentError

from palaestrai.core import RuntimeConfig
from palaestrai.experiment import Executor, ExecutorState
from palaestrai.experiment.experiment_run import ExperimentRun as Experiment


@click.group()
def client():
    pass


@client.command()
@click.argument(
    "experiment_file",
    type=click.Path(
        exists=True, dir_okay=False, readable=True, allow_dash=True
    ),
)
def experiment_start(experiment_file):
    """Start an experiment from a file"""
    experiment = Experiment.load(experiment_file)
    executor = Executor()
    executor.schedule(experiment)
    executor_final_state = asyncio.run(executor.execute())
    if executor_final_state != ExecutorState.EXITED:
        sys.exit(
            {
                ExecutorState.SIGINT: -2,
                ExecutorState.SIGABRT: -6,
                ExecutorState.SIGTERM: -15,
            }[executor_final_state]
        )


def _get_store_uri():
    config = RuntimeConfig()
    if not config.store_uri:
        click.echo(
            "Please create a runtime config (one of %s), "
            "and set the 'store_uri' options.\n"
            "My configuration, loaded from %s, does not contain it."
            % (config.CONFIG_FILE_PATHS, config._config_file_path),
            file=sys.stderr,
        )
        exit(1)

    return config.store_uri


@client.command()
def database_create():
    """Create the store database. Requires a runtime configuration file."""
    store_uri = _get_store_uri()
    if not store_uri:
        exit(1)
    from palaestrai.store.database_util import setup_database

    try:
        setup_database(store_uri)
    except ArgumentError as e:
        click.echo(
            "SQLalchemy could not open a database connection. "
            "Please make sure that your 'store_uri' in %s is formed like "
            "'driver://user:password@host/database'. Error was: %s"
            % (RuntimeConfig()._config_file_path, e),
            file=sys.stderr,
        )
        exit(1)


@client.command()
def database_migrate():
    store_uri = _get_store_uri()
    if not store_uri:
        exit(1)


@client.command()
def runtime_config_show_default():
    """Shows the default runtime config and search paths."""
    yml = ruamel.yaml.YAML(typ="safe")
    print("%YAML 1.2\n")
    print("# palaestrAI Runtime Configuration")
    print(
        f"# Search path for this file is: {RuntimeConfig().CONFIG_FILE_PATHS}"
    )
    print("#\n\n---")
    yml.dump(RuntimeConfig().DEFAULT_CONFIG, stream=sys.stdout)


@client.command()
def runtime_config_show_effective():
    """Shows the currently effective runtime configuration"""
    yml = ruamel.yaml.YAML(typ="safe")
    print("%YAML 1.2")
    print(f"# Configuration loaded from: {RuntimeConfig()._config_file_path}")
    print("---")
    yml.dump(RuntimeConfig().to_dict(), stream=sys.stdout)



"""This module contains the class :class:`Agent` that
stores all information regarding a specific agent.

"""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import List
    from . import Brain, Muscle, SensorInformation, ActuatorInformation


@dataclass
class Agent:
    """Stores information about an agent.

    The agent class is used to store information about an
    agent. It is currently used by the simulation controller
    to have a internal representation of all agents.

    Parameters
    ----------
    uid : uuid4
        The uid is used to identify an agent
    brain: :class:`~.Brain`
        An instance of a palaestrai brain. It
        defines what type of AI is used
    brain_params: dict
        This dictionary contains all parameters needed
        by the brain.
    muscle: :class:`~.Muscle`
        An instance of a palaestrai muscle. It
        defines what type of AI is used and is linked
        to the type of brain
    muscle_params: dict
        This dictionary contains all parameters needed
        by the muscle.
    sensors: any
        The list of sensors the agent is allowed to
        access.
    actuators: any
        The list of actuators the agent is allowed to
        access.
    """

    uid: str
    brain: Brain
    brain_params: dict
    muscle: Muscle
    muscle_params: dict
    sensors: List[SensorInformation]
    actuators: List[ActuatorInformation]



from __future__ import annotations

from dataclasses import dataclass
from typing import List, TYPE_CHECKING

if TYPE_CHECKING:
    from palaestrai.agent import (
        SensorInformation,
        ActuatorInformation,
        RewardInformation,
    )
    from palaestrai.types.mode import Mode


@dataclass
class AgentUpdateRequest:
    """Provides fresh data from a :class:`SimulationController` to
    an :class:`Agent`.

    * Sender: :class:`SimulationController`
    * Receiver: :class:`Muscle`

    :param sender_simulation_controller: The sending
        :class:`SimulationController`
    :param receiver_agent_id: The receiving agent, e.g., a :class:`Muscle`
    :param experiment_run_id: ID of the experiment run this update is part of
    :param actuators: List of actuators available for the agent
    :param sensors: Sensor input data for the agent
    :param reward: Current reward from the environment
    :param is_terminal: Indicates whether this is the last update from the
        environment or not
    """

    sender_simulation_controller_id: str
    receiver_agent_id: str
    experiment_run_id: str
    actuators: List[ActuatorInformation]
    sensors: List[SensorInformation]
    reward: List[RewardInformation]
    is_terminal: bool
    mode: Mode

    @property
    def sender(self):
        return self.sender_simulation_controller_id

    @property
    def receiver(self):
        return self.receiver_agent_id



from typing import TYPE_CHECKING

from palaestrai.core.protocol import (
    NextPhaseRequest,
    SimulationControllerTerminationRequest,
    SimulationControllerTerminationResponse,
)

from . import LOG, RunGovernorState

if TYPE_CHECKING:
    from palaestrai.experiment import RunGovernor, ExperimentRun


class RGSHandlingSimControllerTermination(RunGovernorState):
    """Represents the HANDLING_SIM_CONTROLLER_TERMINATION state.

    This state is entered either after the run governor received a
    SimControllerTerminationRequest during the EXPERIMENT_RUNNING
    state.


    Possible next states are:
    * :class:`.RGSStoppingTransceiving` in the case of a complete shutdown.
    * :class:`.RGSTransceiving` in the default case.

    Parameters
    ----------
    ctx: :class:`.RunGovernor`
        The run governor instance that provides the context for this
        state.
    request: :class:`.SimControllerTerminationRequest`
        The sim controller termination request that leads to this
        state.

    """

    def __init__(self, rgc: "RunGovernor"):
        super().__init__(rgc, "HANDLING_SIM_CONTROLLER_TERMINATION")

    async def run(self):
        request = self.rgc.last_request.pop()
        restart = self._check_episodes()
        is_terminal = self._handle_termination_request(request, restart)

        await self._prepare_reply(request, is_terminal, restart)

    def next_state(self):
        from . import RGSStoppingSimulation, RGSTransceiving

        reply = self.rgc.next_reply[0]

        if isinstance(reply, SimulationControllerTerminationResponse):
            if self.rgc.next_reply[0].complete_shutdown:
                self.rgc.next_reply.pop()
                self.rgc.state = RGSStoppingSimulation(self.rgc)
            else:
                self.rgc.state = RGSTransceiving(self.rgc)
        else:
            self.rgc.next_reply.pop()
            self.rgc.state = RGSStoppingSimulation(self.rgc)

    def _check_episodes(self):

        self.rgc.current_episode += 1
        episodes = self.rgc.experiment_run.get_episodes(self.rgc.current_phase)
        LOG.debug(
            "RunGovernor(id=0x%x, uid=%s) checking episodes: %d/%d.",
            id(self.rgc),
            self.rgc.uid,
            self.rgc.current_episode,
            episodes,
        )
        return self.rgc.current_episode < episodes

    def _handle_termination_request(
        self, request: SimulationControllerTerminationRequest, restart: bool
    ) -> bool:
        LOG.debug(
            "RunGovernor(id=0x%x, uid=%s) handling "
            "SimulationControllerTerminationRequest: %s.",
            id(self.rgc),
            self.rgc.uid,
            request,
        )

        active_sim_controllers = self.rgc.sim_controllers.copy()
        if not restart:
            del active_sim_controllers[request.sender]

        return self.rgc.termination_condition.check_termination(
            request, active_sim_controllers
        )

    async def _prepare_reply(
        self,
        request: SimulationControllerTerminationRequest,
        is_terminal: bool,
        restart: bool,
    ):
        self.rgc.next_reply.append(
            SimulationControllerTerminationResponse(
                sender_run_governor_id=self.rgc.uid,
                receiver_simulation_controller_id=request.sender,
                experiment_run_id=self.rgc.experiment_run_id,
                restart=restart,
                complete_shutdown=is_terminal,
            )
        )
        if is_terminal:
            # All simulation controllers finished
            if self.rgc.experiment_run is None:
                # Should NEVER happen, but mypy goes enrage otherwise
                return

            run: ExperimentRun = self.rgc.experiment_run

            next_phase = self.rgc.current_phase + 1
            if run.has_next_phase(self.rgc.current_phase):
                LOG.info(
                    "RunGovernor(id=0x%x, uid=%s) finished phase %d/%d. "
                    "Shutting down current phase and continue with the "
                    "next phase.",
                    id(self.rgc),
                    self.rgc.uid,
                    next_phase,
                    run.num_phases,
                )
                self.rgc.last_request.append(
                    NextPhaseRequest(
                        sender_run_governor_id=self.rgc.uid,
                        receiver_run_governor_id=self.rgc.uid,
                        next_phase=next_phase,
                    )
                )
            else:
                LOG.info(
                    "RunGovernor(id=0x%x, uid=%s) finished phase %d/%d. "
                    "Reached a complete shutdown.",
                    id(self.rgc),
                    self.rgc.uid,
                    next_phase,
                    run.num_phases,
                )
            await self.rgc.major_domo_worker.transceive(
                self.rgc.next_reply[0],
                True,
            )



from unittest import TestCase
from palaestrai.types import Box, Space
import numpy as np


class BoxTest(TestCase):
    def setUp(self) -> None:
        self.box_float = Box(
            low=-1.0, high=2.0, shape=(3, 4), dtype=np.float32
        )
        self.box_float_inner = Box(
            low=0.0, high=1.0, shape=(3, 4), dtype=np.float32
        )
        self.box_diff_bounds = Box(
            low=np.array([0.0, -10.0, 12]),
            high=np.array([1.0, 10.0, 78]),
            dtype=np.float32,
        )
        self.box_diff_bounds_arrays = Box(
            low=[[0.0, 1.0], [2.0, 3.0]],
            high=[[0.0, 1.0], [2.0, 3.0]],
            dtype=np.float32,
        )

    def test_sample(self):
        for x in range(0, 50):
            self.assertTrue(
                self.box_float.sample().min() > -1
                and self.box_float.sample().max() < 3
            )

    def test_contains(self):
        # box should contain its own samples
        sample_self = self.box_float.sample()
        self.assertTrue(self.box_float.contains(sample_self))

        # box should contain samples of more restricted boxes
        sample_other = self.box_float_inner.sample()
        self.assertTrue(self.box_float.contains(sample_other))

        # all values inside of box
        array_expected_true = [
            [0.41128066, 0.3421888, 1.2279652, 1.2973598],
            [1.7220176, 0.7265164, 1.591186, 1.2400235],
            [1.9035902, -0.62997127, 0.12461133, 0.02076343],
        ]
        self.assertTrue(self.box_float.contains(array_expected_true))

        # array not in box because 25.411 > 2.0
        array_expected_false = [
            [25, 0.3421888, 1.2279652, 1.2973598],
            [1.7220176, 0.7265164, 1.591186, 1.2400235],
            [1.9035902, -0.62997127, 0.12461133, 0.02076343],
        ]
        self.assertFalse(self.box_float.contains(array_expected_false))

        # lower and upper bound inside of box
        lower_bound = [[-1, -1, -1, -1]] * 3
        self.assertTrue(self.box_float.contains(lower_bound))
        upper_bound = [[2, 2, 2, 2]] * 3
        self.assertTrue(self.box_float.contains(upper_bound))

    def test_contains_single_value(self):
        b = Box(low=0.0, high=1.0, shape=(1,), dtype=np.float32)
        self.assertTrue(b.contains(float(0.52620125)))

    def test_contains_single_value_in_list(self):
        b = Box(
            low=np.array([0.0]),
            high=np.array([1.0]),
            shape=(1,),
            dtype=np.float32,
        )
        self.assertTrue(b.contains([0.4564]))

    def test_contains_values_list(self):
        b = Box(low=0.0, high=1.1, shape=(3,), dtype=np.float32)
        self.assertTrue(b.contains([0.4564, 0.12314, 0.738457]))

    def test_contains_values_array(self):
        b = Box(
            low=[[0.0, 1.0], [2.0, 3.0]],
            high=[[1.0, 2.0], [3.0, 4.0]],
            dtype=np.float32,
        )
        self.assertTrue(b.contains([[0.12314, 1.738457], [2.3020, 3.0123]]))

    def test_to_string_equals(self):
        self.assertEqual(
            Box(low=-1.0, high=1.0, shape=(1,)).to_string(),
            "Box(low=[-1.0], high=[1.0], shape=(1,), dtype=np.float32)",
        )

    def test_from_to_string(self):
        box1 = Box.from_string(
            "Box(low=[  0., -10.,  12.], "
            "   high=[ 1., 10., 78.], dtype=np.float32)"
        )
        self.assertEqual(box1, Box.from_string(box1.to_string()))
        self.assertEqual(box1, Space.from_string(box1.to_string()))

    def test_from_string(self):
        self.assertEqual(
            self.box_float,
            Box.from_string(
                "Box(low=-1.0, high=2.0," "shape=(3, 4), dtype=np.float32)"
            ),
        )
        self.assertEqual(
            self.box_float,
            Box.from_string(
                " Box(low=-1.0, high=2.0, shape=(3, 4), dtype=np.float32)"
            ),
        )
        self.assertEqual(
            self.box_diff_bounds,
            Box.from_string(
                "Box(low=[  0., -10.,  12.],"
                "high=[ 1., 10., 78.],"
                "dtype=np.float32)"
            ),
        )

        self.assertEqual(
            self.box_diff_bounds_arrays,
            Box.from_string(
                "Box(low=[ [ 0., 1.], [ 2., 3.]],"
                "high=[ [0., 1.], [2., 3.]],"
                "dtype=np.float32)"
            ),
        )

    def test_space_factory(self):
        self.assertEqual(
            self.box_float,
            Space.from_string(
                " Box(low=-1.0, high=2.0, shape=(3, 4), dtype=np.float32)"
            ),
        )

    def test_to_string_irregular(self):
        box_irr = (
            "Box(low=[-1.0, -1.0, -1.0], "
            "high=[1.0, 1.0, 1.0], "
            "shape=(3,), "
            "dtype=np.float32)"
        )
        box1 = Box(low=-1, high=1, shape=(3,))
        self.assertEqual(box_irr, box1.to_string())

    def test_string_compatibility(self):
        box = self.box_diff_bounds.to_string()
        diff_bounds_fs = Space.from_string(box)
        self.assertEqual(diff_bounds_fs, self.box_diff_bounds)

        box = self.box_float.to_string()
        float_fs = Box.from_string(box)
        self.assertEqual(float_fs, self.box_float)



import asyncio
import queue
import signal
import time
from binascii import hexlify
from copy import copy
from typing import List, Dict

import zmq
import zmq.asyncio
from zmq.asyncio import Socket, Poller
import palaestrai.core.MDP as MDP
from palaestrai.store.receiver import StoreReceiver
from . import LOG
from .zhelpers import pprint_message, ContextGuard


class Service(object):
    """A single service managed by this broker"""

    def __init__(self, name):
        """Creates a new Service object for a given service name

        :param name: The service name
        """
        self.name = name
        self.requests = []
        self.waiting = []

    def __str__(self):
        return "Service(name=%s, requests=%s, waiting=%s)" % (
            self.name,
            self.requests,
            self.waiting,
        )


class Worker(object):
    """Represents an external Worker in the major domo protocol.

    The major domo protocol knows clients (who send out requests) and workers,
    who process these requests. This internal data structure represents such
    an external worker and is used for bookkeeping.
    """

    def __init__(self, identity, address, lifetime):
        """Create a new worker structure

        :param identity: The hex identity of the worker; can be arbitrary, but
            must be unique
        :param address: The workers TCP connection address
        :param lifetime: How long the worker can stay in the books
        """
        self.service = None
        self.address = address
        self.identity = identity
        self.expiry = time.time() + 1e-3 * lifetime

    def __str__(self):
        return "Worker(identity=%s, address=%s, service=%s, expiry=%s)" % (
            self.identity,
            self.address,
            self.service,
            self.expiry,
        )

    def __repr__(self):
        return str(self)


class MajorDomoBroker:
    """Distributes messages between clients and workers according to services

    This Major Domo Protocol broker is a minimal implementation of
    http:#rfc.zeromq.org/spec:7 and spec:8.
    """

    INTERNAL_SERVICE_PREFIX = b"mmi."
    HEARTBEAT_LIVENESS = 3  # 3-5 is reasonable
    HEARTBEAT_INTERVAL = 2500  # msecs
    HEARTBEAT_EXPIRY = HEARTBEAT_INTERVAL * HEARTBEAT_LIVENESS

    def __init__(self, uri=None):
        """Create a new MajorDomoBroker to listen at a specific URI

        :param uri: The URI the broker should listen on
        """
        self._running: bool = True

        self._uri: str = uri
        self._socket: Socket = None
        self._poller: Poller = None
        self._services: Dict[Service] = {}
        self._workers: Dict[Worker] = {}
        self._waiting: List[Worker] = []
        self._heartbeat_at = (
            time.time() + 1e-3 * MajorDomoBroker.HEARTBEAT_INTERVAL
        )

        self._queue: queue.Queue = None
        self._store_receiver: StoreReceiver = None

        LOG.debug("New MajorDomoBroker(id=0x%x, uri=%s)", id(self), self._uri)

    def _handle_signal_interrupt(self, signum):
        if signum in {signal.SIGABRT, signal.SIGTERM}:
            LOG.debug(
                "MajorDomoBroker(id=0x%x, uri=%s) "
                "received signal %s, terminating",
                id(self),
                self._uri,
                signum,
            )
            self._running = False
        elif signum == signal.SIGINT:
            LOG.debug(
                "MajorDomoBroker(id=0x%x, uri=%s) "
                "staying online after SIGINT for as long as possible "
                "to allow for an orderly shutdown",
                id(self),
                self._uri,
            )
        else:
            LOG.debug(
                "MajorDomoBroker(id=0x%x, uri=%s) "
                "received signal %s, but will ignore it",
                id(self),
                self._uri,
                signum,
            )

    def _init_signal_handler(self):
        signals = {signal.SIGINT, signal.SIGABRT, signal.SIGTERM}
        LOG.debug(
            "MajorDomoBroker(id=0x%x) "
            "registering signal handlers for signals %s",
            id(self),
            signals,
        )
        loop = asyncio.get_running_loop()
        for signum in signals:
            loop.add_signal_handler(
                signum, self._handle_signal_interrupt, signum
            )

    def _init_communication(self):
        self._socket = zmq.asyncio.Socket(
            context=ContextGuard.asyncio_context(), socket_type=zmq.ROUTER
        )  # self.ctx.socket(zmq.ROUTER)
        self._socket.linger = 0
        self._poller = zmq.asyncio.Poller()
        self._poller.register(self._socket, zmq.POLLIN)
        self.bind(self._uri)

    def _init_store(self):
        self._queue = queue.Queue()
        self._store_receiver = StoreReceiver(self._queue)
        self._store_receiver.start()

    async def mediate(self):
        """Mediation loop for message distribution

        This method is an infinite loop that receives and distributes messages
        received from clients and workers.
        """
        LOG.debug(
            "MajorDomoBroker(id=0x%x, uri=%s) "
            "MDP broker/0.1.1 starting mediation",
            id(self),
            self._uri,
        )

        self._init_signal_handler()
        self._init_communication()
        self._init_store()

        LOG.debug(
            "MajorDomoBroker(id=0x%x, uri=%s) "
            "started StoreReceiver(id=0x%x, uid=%s)",
            id(self),
            self._uri,
            id(self._store_receiver),
            self._store_receiver.uid,
        )

        while self._running:
            items = await self._poller.poll(self.HEARTBEAT_INTERVAL)
            if not items or not self._running:
                continue
            msg = await self._socket.recv_multipart()
            msg_dup = copy(msg)
            sender = msg.pop(0)
            empty = msg.pop(0)
            assert empty == b""
            header = msg.pop(0)

            LOG.debug(
                "MajorDomoBroker(id=0x%d, uri=%s) received message "
                "from '%s' of type '%s'",
                id(self),
                self._uri,
                sender,
                header,
            )

            if MDP.C_CLIENT == header:
                if msg[0] == MDP.W_DESTROY:
                    await self._destroy()
                    break
                await self.process_client(sender, msg)
            elif MDP.W_WORKER == header:
                await self._process_worker(sender, msg)
            else:
                LOG.debug(
                    "MajorDomoBroker(id=0x%d, uri=%s) "
                    "received invalid message, ignoring",
                    id(self),
                    self._uri,
                )
            await self.purge_workers()
            await self.send_heartbeats()
            self._queue.put_nowait(msg_dup)  # Process for store.
        self._store_receiver.shutdown()
        self._store_receiver.join()

    async def _destroy(self):
        """Disconnect all workers, destroy context."""
        LOG.debug(
            "MajorDomoBroker(id=0x%d, uri=%s) " "destroying workers",
            id(self),
            self._uri,
        )
        while self._workers:
            await self._delete_worker(list(self._workers.values())[0], True)

    async def process_client(self, sender, msg):
        """Process a request coming from a client."""
        assert len(msg) >= 2  # Service name + body
        service = msg.pop(0)
        # Set reply return address to client sender
        msg = [sender, b""] + msg
        if service.startswith(self.INTERNAL_SERVICE_PREFIX):
            await self.service_internal(service, msg)
        else:
            await self.dispatch(self._require_service(service), msg)

    async def _process_worker(self, sender, msg):
        """Process message sent to us by a worker."""
        assert len(msg) >= 1  # At least, command
        command = msg.pop(0)

        worker_ready = hexlify(sender) in self._workers
        worker = self.require_worker(sender)

        if MDP.W_READY == command:
            assert len(msg) >= 1  # At least, a service name
            service = msg.pop(0)
            # Not first command in session or Reserved service name
            if worker_ready or service.startswith(
                self.INTERNAL_SERVICE_PREFIX
            ):
                await self._delete_worker(worker, True)
            else:
                # Attach worker to service and mark as idle
                worker.service = self._require_service(service)
                await self.worker_waiting(worker)

        elif MDP.W_REPLY == command:
            if worker_ready:
                # Remove & save client return envelope and insert the
                # protocol header and service name, then rewrap envelope.
                client = msg.pop(0)
                _ = msg.pop(0)  # Empty delimiter frame, see MDP definition
                msg = [client, b"", MDP.C_CLIENT, worker.service.name] + msg
                await self._socket.send_multipart(msg)
                await self.worker_waiting(worker)
            else:
                await self._delete_worker(worker, True)

        elif MDP.W_HEARTBEAT == command:
            if worker_ready:
                worker.expiry = time.time() + 1e-3 * self.HEARTBEAT_EXPIRY
            else:
                await self._delete_worker(worker, True)

        elif MDP.W_DISCONNECT == command:
            await self._delete_worker(worker, False)
        else:
            LOG.error("Broker received invalid message")
            LOG.debug(pprint_message(msg))

    async def _delete_worker(self, worker, disconnect):
        """Deletes worker from all data structures, and deletes worker."""
        assert worker is not None
        if disconnect:
            await self.send_to_worker(worker, MDP.W_DISCONNECT, None, None)

        if worker.service is not None and worker in worker.service.waiting:
            worker.service.waiting.remove(worker)

        if worker.identity not in self._workers:
            LOG.ERROR(
                "Worker identity '%s' is missing in workers %s (dc: %s).",
                worker.identity,
                self._workers,
                str(disconnect),
            )
        self._workers.pop(worker.identity)

    def require_worker(self, address) -> Worker:
        """Finds the worker (creates if necessary)."""
        assert address is not None
        identity = hexlify(address)
        worker = self._workers.get(identity)
        if worker:
            return worker
        worker = Worker(identity, address, self.HEARTBEAT_EXPIRY)
        self._workers[identity] = worker
        LOG.debug(
            "MajorDomoBroker(id=0x%x, uri=%s) "
            "registered new Worker(identity=%s, address=%s)",
            id(self),
            self._uri,
            identity,
            address,
        )
        return worker

    def _require_service(self, name):
        """Locates the service (creates if necessary)."""
        assert name is not None
        service = self._services.get(name)
        if service:
            return service
        service = Service(name)
        self._services[name] = service
        return service

    def bind(self, endpoint):
        """Bind broker to endpoint, can call this multiple times.

        We use a single socket for both clients and workers.
        """
        try:
            self._socket.bind(endpoint)
        except zmq.error.ZMQError as e:
            LOG.fatal(
                "MajorDomoBroker(id=0x%x, uri=%s) " "failed to bind: %s",
                id(self),
                self._uri,
                e,
            )
            raise e
        LOG.debug(
            "MajorDomoBroker(id=0x%x, uri=%s) bound successfully to socket",
            id(self),
            endpoint,
        )

    async def service_internal(self, service, msg):
        """Handle internal service according to 8/MMI specification"""
        returncode = b"501"
        if b"mmi.service" == service:
            name = msg[-1]
            returncode = b"200" if name in self._services else b"404"
        msg[-1] = returncode

        # insert the protocol header and service name
        # after the routing envelope ([client, '']):
        msg = msg[:2] + [MDP.C_CLIENT, service] + msg[2:]
        await self._socket.send_multipart(msg)

    async def send_heartbeats(self):
        """Send heartbeats to idle workers if neccessary.

        This method checks for the current time elapsed being past the next
        designated checkpoint, and if yes, sends a heartbeat message to all
        workers.
        """
        if time.time() <= self._heartbeat_at:
            return
        for worker in self._waiting:
            LOG.debug("Sending heartbeat to worker %s", worker)
            await self.send_to_worker(worker, MDP.W_HEARTBEAT, None, None)
        self._heartbeat_at = time.time() + 1e-3 * self.HEARTBEAT_INTERVAL

    async def purge_workers(self):
        """Look for & kill expired workers."""
        while self._waiting:
            # Workers are stored from oldest to most recent, so we pop
            # until we find an active one.
            w = self._waiting[0]
            if w.expiry >= time.time():
                break
            LOG.info("Broker deleting expired worker: %s", w.identity)
            await self._delete_worker(w, False)
            self._waiting.pop(0)

    async def worker_waiting(self, worker):
        """This worker is now waiting for work."""
        self._waiting.append(worker)
        worker.service.waiting.append(worker)
        worker.expiry = time.time() + 1e-3 * self.HEARTBEAT_EXPIRY
        await self.dispatch(worker.service, None)

    async def dispatch(self, service, msg):
        """Dispatch requests to waiting workers as possible"""
        assert service is not None
        if msg is not None:  # Queue message if any
            service.requests.append(msg)
        await self.purge_workers()
        LOG.debug(
            "MajorDomoBroker(id=0x%x, uri=%s) "
            "is dispatching a message to service '%s'; "
            "Services: %s %s, Workers: %s %s",
            id(self),
            self._uri,
            service,
            len(self._services.keys()),
            self._services.keys(),
            len(self._workers.keys()),
            self._workers,
        )
        while service.waiting and service.requests:
            msg = service.requests.pop(0)
            worker = service.waiting.pop(0)
            self._waiting.remove(worker)
            await self.send_to_worker(worker, MDP.W_REQUEST, None, msg)

    async def send_to_worker(self, worker, command, option, msg=None):
        """Send message to worker.

        If message is provided, sends that message.
        """
        if msg is None:
            msg = []
        elif not isinstance(msg, list):
            msg = [msg]

        # Stack routing and protocol envelopes to start of message
        # and routing envelope
        if option is not None:
            msg = [option] + msg
        msg = [worker.address, b"", MDP.W_WORKER, command] + msg
        LOG.debug(
            "MajorDomoBroker(id=0x%x, uri=%s) "
            "sending message with command '%r' to '%s'",
            id(self),
            self._uri,
            command,
            worker,
        )
        await self._socket.send_multipart(msg)



import re
import ast
import numpy as np
import numbers
from .space import Space
from palaestrai.util.dynaloader import locate


class Box(Space):
    """A box in R^n, i.e.each coordinate is bounded.

    There are two common use cases:

    * Identical bound for each dimension::
        >>> Box(low=-1.0, high=2.0, shape=(3, 4), dtype=np.float32)
        Box(3, 4)

    * Independent bound for each dimension::
        >>> Box(low=np.array([-1.0, -2.0]), high=np.array([2.0, 4.0]), dtype=np.float32)
        Box(2,)

    """

    _BOX_RE = re.compile(r"\A\s*?Box\((.+)\)\s*\Z")
    _SHAPE_RE = re.compile(r"shape=\(((\d+, ?)|((\d+, ?)+\d+ ?))\)")
    _DTYPE_RE = re.compile(r"dtype=(np\.)?(\w+)")
    _LOW_BOUNDARY_RE = re.compile(
        r"low=((?P<single>[0-9.-]+)|(?P<list>\[[^a-zA-Z]*\]))"
    )
    _HIGH_BOUNDARY_RE = re.compile(
        r"high=((?P<single>[0-9.-]+)|(?P<list>\[[^a-zA-Z]*\]))"
    )

    def __init__(self, low, high, shape=None, dtype=np.float32):
        assert dtype is not None, "dtype must be explicitly provided. "
        self.dtype = np.dtype(dtype)

        if shape is None:
            if np.array(low).shape != np.array(low).shape:
                raise ValueError("low and high shapes are not compatible")
            self.shape = np.array(low).shape
            self.low = low
            self.high = high
        else:
            self.shape = tuple(shape)

        if isinstance(low, numbers.Number) and isinstance(
            high, numbers.Number
        ):
            self.low = np.array(low)
            self.high = np.array(high)
            self.shape = tuple(shape)
            self.low = np.full(self.shape, low)
            self.high = np.full(self.shape, high)
            self.low = low + np.zeros(shape)
            self.high = high + np.zeros(shape)
        else:
            if hasattr(low, "__iter__"):
                self.low = np.array(low)
            else:
                raise ValueError(
                    "Incompatible type '%s' for lower boundary" % type(low)
                )
            if hasattr(high, "__iter__"):
                self.high = np.array(high)
            else:
                raise ValueError(
                    "Incompatible type '%s' for higher boundary" % type(high)
                )

        self.low = self.low.astype(self.dtype)
        self.high = self.high.astype(self.dtype)
        super(Box, self).__init__(self.shape, self.dtype)

    def sample(self):
        high = (
            self.high
            if self.dtype.kind == "f"
            else self.high.astype("int64") + 1
        )
        return self.np_random.uniform(
            low=self.low, high=high, size=self.shape
        ).astype(self.dtype)

    def contains(self, x):
        if (
            isinstance(x, np.float32)
            or isinstance(x, int)
            or isinstance(x, float)
        ):
            x = np.array([x])
        if isinstance(x, list):
            x = np.array(x)  # Promote list to array for contains check
        return (
            x.shape == self.shape
            and np.all(x >= self.low)
            and np.all(x <= self.high)
        )

    def to_string(self):
        return "Box(low=%s, high=%s, shape=%s, dtype=np.%s)" % (
            self.low.tolist(),
            self.high.tolist(),
            self.shape,
            self.dtype,
        )

    @classmethod
    def from_string(cls, s):
        complete_match = Box._BOX_RE.match(s)
        if not complete_match:
            raise RuntimeError(
                "String '%s' does not match '%s'" % (s, Box._BOX_RE)
            )
        inner_str = complete_match[1]

        list_split_re = re.compile(r",\s*|\s+")

        lower_boundary_match = Box._LOW_BOUNDARY_RE.search(inner_str)
        if not lower_boundary_match:
            raise RuntimeError(
                "No lower boundary in '%s' (%s)" % (s, Box._LOW_BOUNDARY_RE)
            )
        lower = None
        if lower_boundary_match["single"]:
            lower = float(lower_boundary_match["single"])
        else:  # list
            lower = np.array(ast.literal_eval(lower_boundary_match["list"]))

        higher_boundary_match = Box._HIGH_BOUNDARY_RE.search(inner_str)
        if not higher_boundary_match:
            raise RuntimeError(
                "No higher boundary in '%s' (%s)" % (s, Box._HIGH_BOUNDARY_RE)
            )
        higher = None
        if higher_boundary_match["single"]:
            higher = float(higher_boundary_match["single"])
        else:  # list
            higher = np.array(ast.literal_eval(higher_boundary_match["list"]))

        shape = None
        if not type(higher) is np.ndarray:
            shape_match = Box._SHAPE_RE.search(inner_str)
            if not shape_match:
                raise RuntimeError(
                    "No or invalid shape in '%s' ('%s')"
                    % (inner_str, Box._SHAPE_RE)
                )
            shape = tuple(
                [
                    int(i)
                    for i in shape_match[1].replace(" ", "").split(",")
                    if i
                ]
            )

        dtype_match = Box._DTYPE_RE.search(inner_str)
        if not dtype_match:
            raise RuntimeError(
                "No or invalid dtype in '%s' (pattern '%s')"
                % (inner_str, Box._DTYPE_RE)
            )
        dtype = locate("numpy.%s" % dtype_match[2])

        return Box(low=lower, high=higher, shape=shape, dtype=dtype)

    def __eq__(self, other):
        return (
            isinstance(other, Box)
            and (self.shape == other.shape)
            and np.allclose(self.low, other.low)
            and np.allclose(self.high, other.high)
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from palaestrai.agent import ActuatorInformation, SensorInformation


@dataclass
class EnvironmentResetResponse:
    """Response to a reset of an :class:`.Environment`.

    Parameters
    ----------
    sender_environment_id: str
        ID of the sending :class:`.Environment`.
    receiver_simulation_controller_id: str
        ID of the receiving :class:`.SimulationController`.
    create_new_instance: bool
        If set to True, the SimulationController will create a new
        instance of the environment.
    sensors: List[SensorInformation]
        List of :class:`.SensorInformation` after the reset. Should
        normally be the same list as after the first start.
    actuators: List[ActuatorInformation]
        List of :class:`.ActuatorInformation` after the reset. Should
        normally be the same list as after the first start.

    """

    sender_environment_id: str
    receiver_simulation_controller_id: str
    create_new_instance: bool
    sensors: List[SensorInformation]
    actuators: List[ActuatorInformation]

    @property
    def sender(self):
        return self.sender_environment_id

    @property
    def receiver(self):
        return self.receiver_simulation_controller_id
